import os
from typing import List, Optional

from anyscale._private.anyscale_client.common import AnyscaleClientInterface
from anyscale._private.models.image_uri import ImageURI
from anyscale._private.sdk.base_sdk import BaseSDK
from anyscale.cli_logger import BlockLogger
from anyscale.sdk.anyscale_client.models.cluster_environment_build import (
    ClusterEnvironmentBuild,
)
from anyscale.sdk.anyscale_client.models.cluster_environment_build_status import (
    ClusterEnvironmentBuildStatus,
)


ENABLE_IMAGE_BUILD_FOR_TRACKED_REQUIREMENTS = (
    os.environ.get("ANYSCALE_ENABLE_IMAGE_BUILD_FOR_TRACKED_REQUIREMENTS", "0") == "1"
)


class ImageSDK(BaseSDK):
    def __init__(
        self,
        *,
        logger: Optional[BlockLogger] = None,
        client: Optional[AnyscaleClientInterface] = None,
    ):
        super().__init__(logger=logger, client=client)
        self._enable_image_build_for_tracked_requirements = (
            ENABLE_IMAGE_BUILD_FOR_TRACKED_REQUIREMENTS
        )

    @property
    def enable_image_build_for_tracked_requirements(self) -> bool:
        return self._enable_image_build_for_tracked_requirements

    def get_default_image(self) -> str:
        return self.client.get_default_build_id()

    def get_image_build(self, build_id: str) -> Optional[ClusterEnvironmentBuild]:
        return self.client.get_cluster_env_build(build_id)

    def build_image_from_containerfile(self, name: str, containerfile: str) -> str:
        return self.client.get_cluster_env_build_id_from_containerfile(
            cluster_env_name=name, containerfile=containerfile, anonymous=False
        )

    def build_image_from_requirements(
        self, name: str, base_build_id: str, requirements: List[str]
    ):
        if requirements:
            base_build = self.client.get_cluster_env_build(base_build_id)
            if (
                base_build
                and base_build.status == ClusterEnvironmentBuildStatus.SUCCEEDED
                and base_build.docker_image_name
            ):
                self.logger.info(f"Using tracked python packages: {requirements}")
                lines = [
                    "# syntax=docker/dockerfile:1",
                    f"FROM {base_build.docker_image_name}",
                ]
                for requirement in requirements:
                    lines.append(f'RUN pip install "{requirement}"')
                return self.build_image_from_containerfile(name, "\n".join(lines))
            else:
                raise RuntimeError(
                    f"Base build {base_build_id} is not a successful build."
                )
        else:
            return base_build_id

    def registery_image(self, image_uri: str) -> str:
        image_uri_checked = ImageURI.from_str(image_uri_str=image_uri)
        return self.client.get_cluster_env_build_id_from_image_uri(
            image_uri=image_uri_checked
        )

    def get_image_uri_from_build_id(self, build_id: str) -> Optional[ImageURI]:
        return self.client.get_cluster_env_build_image_uri(
            cluster_env_build_id=build_id
        )



from typing import Optional

import click

from anyscale.controllers.schedule_controller import ScheduleController


@click.group("schedule", help="Create and manage Anyscale Schedules.")
def schedule_cli() -> None:
    pass


@schedule_cli.command(name="create")
@click.argument("schedule_config_file", required=True)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option(
    "--description", required=False, default=None, help="Description of schedule."
)
def create(
    schedule_config_file: str, name: Optional[str], description: Optional[str],
) -> None:
    """ Create or Update a Schedule

    This function accepts 1 argument, a path to a YAML config file that defines this schedule.

    Note: if a schedule with the name exists in the specified project, it will be updated instead.
    """

    job_controller = ScheduleController()
    job_controller.apply(
        schedule_config_file, name, description,
    )


@schedule_cli.command(name="update")
@click.argument("schedule_config_file", required=True)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option(
    "--description", required=False, default=None, help="Description of schedule."
)
def update(
    schedule_config_file: str, name: Optional[str], description: Optional[str],
) -> None:
    """ Create or Update a Schedule

    This function accepts 1 argument, a path to a YAML config file that defines this schedule.
    """

    job_controller = ScheduleController()
    job_controller.apply(
        schedule_config_file, name, description,
    )


@schedule_cli.command(name="list",)
@click.option(
    "--name", "-n", required=False, default=None, help="Filter by the name of the job"
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
def list(  # noqa: A001
    name: Optional[str] = None, id: Optional[str] = None  # noqa: A002
) -> None:
    """ List Schedules

    You can optionally filter schedules by rowname.
    """
    job_controller = ScheduleController()
    job_controller.list(name=name, id=id)


@schedule_cli.command(name="pause",)
@click.argument("schedule_config_file", required=False)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
def pause(schedule_config_file: str, name: str, id: str) -> None:  # noqa: A002
    """ Pause a Schedule

    This function accepts 1 argument, a path to a YAML config file that defines this schedule.
    You can also specify the schedule by name or id.
    """

    job_controller = ScheduleController()
    id = job_controller.resolve_file_name_or_id(  # noqa: A001
        schedule_config_file=schedule_config_file, id=id, name=name
    )
    job_controller.pause(id, is_paused=True)


@schedule_cli.command(name="resume",)
@click.argument("schedule_config_file", required=False)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
def unpause(schedule_config_file: str, name: str, id: str) -> None:  # noqa: A002
    """ Resume a Schedule

    This function accepts 1 argument, a path to a YAML config file that defines this schedule.
    You can also specify the schedule by name or id.
    """

    job_controller = ScheduleController()
    id = job_controller.resolve_file_name_or_id(  # noqa: A001
        schedule_config_file=schedule_config_file, id=id, name=name
    )
    job_controller.pause(id, is_paused=False)


@schedule_cli.command(name="run",)
@click.argument("schedule_config_file", required=False)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
def trigger(schedule_config_file: str, id: str, name: str) -> None:  # noqa: A002
    """ Manually run a Schedule

    This function takes an existing schedule and runs it now.
    You can specify the schedule by name or id.
    You can also pass in a YAML file as a convinience. This is equivalent to passing in the name specified in the YAML file.
    IMPORTANT: if you pass in a YAML definition that differs from the Schedule defition, the Schedule will NOT be updated.
    Please use the `anyscale schedule update` command to update the configuration of your schedule
    or use the `anyscale job submit` command to submit a one off job that is not a part of a schedule.
    """

    job_controller = ScheduleController()
    id = job_controller.resolve_file_name_or_id(  # noqa: A001
        schedule_config_file=schedule_config_file, id=id, name=name
    )
    job_controller.trigger(id)


@schedule_cli.command(name="url",)
@click.argument("schedule_config_file", required=False)
@click.option(
    "--name", "-n", required=False, default=None, help="Name of the schedule."
)
@click.option("--id", "-i", required=False, default=None, help="Id of the schedule.")
def url(schedule_config_file: str, id: str, name: str) -> None:  # noqa: A002
    """ Get a Schedule URL

    This function accepts 1 argument, a path to a YAML config file that defines this schedule.
    You can also specify the schedule by name or id.
    """

    job_controller = ScheduleController()
    id = job_controller.resolve_file_name_or_id(  # noqa: A001
        schedule_config_file=schedule_config_file, id=id, name=name
    )
    job_controller.url(id)



from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Union

from anyscale._private.sdk.base_sdk import BaseSDK
from anyscale.client.openapi_client.models import (
    Cloud,
    CloudProviders,
    ComputeNodeType,
    ComputeTemplateConfig,
    DecoratedComputeTemplate,
    DecoratedComputeTemplateConfig,
    Resources,
    WorkerNodeType,
)
from anyscale.cluster_compute import parse_cluster_compute_name_version
from anyscale.compute_config.models import (
    ComputeConfig,
    ComputeConfigVersion,
    HeadNodeConfig,
    MarketType,
    WorkerNodeGroupConfig,
)
from anyscale.sdk.anyscale_client.models import ClusterComputeConfig


# Used to explicitly make the head node unschedulable.
# We can't leave resources empty because the backend will fill in CPU and GPU
# to match the instance type hardware.
UNSCHEDULABLE_RESOURCES = Resources(cpu=0, gpu=0)


class ComputeConfigSDK(BaseSDK):
    def _populate_advanced_instance_config(
        self,
        config: Union[ComputeConfig, HeadNodeConfig, WorkerNodeGroupConfig],
        api_model: Union[ComputeTemplateConfig, ComputeNodeType, WorkerNodeType],
        *,
        cloud: Cloud,
    ):
        """Populates the appropriate advanced instance config field of the API model in place."""
        if not config.advanced_instance_config:
            return

        if cloud.provider == CloudProviders.AWS:
            api_model.aws_advanced_configurations_json = config.advanced_instance_config
        elif cloud.provider == CloudProviders.GCP:
            api_model.gcp_advanced_configurations_json = config.advanced_instance_config
        else:
            raise ValueError(
                f"'advanced_instance_config' is not implemented for cloud provider: '{cloud.provider}'"
            )

    def _convert_resource_dict_to_api_model(
        self, resource_dict: Optional[Dict[str, float]]
    ) -> Optional[Resources]:
        if resource_dict is None:
            return None

        resource_dict = deepcopy(resource_dict)
        return Resources(
            cpu=resource_dict.pop("CPU", None),
            gpu=resource_dict.pop("GPU", None),
            memory=resource_dict.pop("memory", None),
            object_store_memory=resource_dict.pop("object_store_memory", None),
            custom_resources=resource_dict or None,
        )

    def _convert_head_node_config_to_api_model(
        self,
        config: Union[None, Dict, HeadNodeConfig],
        *,
        cloud: Cloud,
        schedulable_by_default: bool,
    ) -> ComputeNodeType:
        if config is None:
            # If no head node config is provided, use the cloud default.
            default: ClusterComputeConfig = self._client.get_default_compute_config(
                cloud_id=cloud.id
            ).config

            api_model = ComputeNodeType(
                name="head-node",
                instance_type=default.head_node_type.instance_type,
                # Let the backend populate the physical resources
                # (regardless of what the default compute config says).
                resources=None if schedulable_by_default else UNSCHEDULABLE_RESOURCES,
            )
        else:
            # Make mypy happy.
            assert isinstance(config, HeadNodeConfig)

            api_model = ComputeNodeType(
                name="head-node",
                instance_type=config.instance_type,
                resources=self._convert_resource_dict_to_api_model(config.resources)
                if config.resources is not None or schedulable_by_default
                else UNSCHEDULABLE_RESOURCES,
            )
            self._populate_advanced_instance_config(
                config, api_model, cloud=cloud,
            )

        return api_model

    def _convert_worker_node_group_configs_to_api_models(
        self,
        configs: Optional[List[Union[Dict, WorkerNodeGroupConfig]]],
        *,
        cloud: Cloud,
    ) -> Optional[List[WorkerNodeType]]:
        if configs is None:
            return None

        api_models = []
        for config in configs:
            # Make mypy happy.
            assert isinstance(config, WorkerNodeGroupConfig)

            api_model = WorkerNodeType(
                name=config.name,
                instance_type=config.instance_type,
                resources=self._convert_resource_dict_to_api_model(config.resources),
                min_workers=config.min_nodes,
                max_workers=config.max_nodes,
                use_spot=config.market_type
                in {MarketType.SPOT, MarketType.PREFER_SPOT},
                fallback_to_ondemand=config.market_type == MarketType.PREFER_SPOT,
            )
            self._populate_advanced_instance_config(
                config, api_model, cloud=cloud,
            )
            api_models.append(api_model)

        return api_models

    def _convert_compute_config_to_api_model(
        self, compute_config: ComputeConfig
    ) -> ComputeTemplateConfig:
        # We should only make the head node schedulable when it's the *only* node in the cluster.
        # `worker_nodes=None` uses the default serverless config, so this only happens if `worker_nodes`
        # is explicitly set to an empty list.
        # Returns the default cloud if user-provided cloud is not specified (`None`).
        cloud_id = self.client.get_cloud_id(cloud_name=compute_config.cloud)
        cloud = self.client.get_cloud(cloud_id=cloud_id)
        if cloud is None:
            raise RuntimeError(
                f"Cloud with ID '{cloud_id}' not found. "
                "This should never happen; please reach out to Anyscale support."
            )

        flags: Dict[str, Any] = {
            "allow-cross-zone-autoscaling": compute_config.enable_cross_zone_scaling,
        }

        # NOTE(edoakes): we currently only support CPU and GPU in `max_resources`.
        # This is validated in the model.
        if compute_config.max_resources:
            if "CPU" in compute_config.max_resources:
                flags["max-cpus"] = compute_config.max_resources["CPU"]
            if "GPU" in compute_config.max_resources:
                flags["max-gpus"] = compute_config.max_resources["GPU"]

        api_model = ComputeTemplateConfig(
            cloud_id=cloud_id,
            allowed_azs=compute_config.zones,
            region="",
            head_node_type=self._convert_head_node_config_to_api_model(
                compute_config.head_node,
                cloud=cloud,
                schedulable_by_default=compute_config.worker_nodes == [],
            ),
            worker_node_types=self._convert_worker_node_group_configs_to_api_models(
                compute_config.worker_nodes, cloud=cloud,
            ),
            auto_select_worker_config=compute_config.worker_nodes is None,
            flags=flags,
        )
        self._populate_advanced_instance_config(
            compute_config, api_model, cloud=cloud,
        )
        return api_model

    def create_compute_config(
        self, compute_config: ComputeConfig, *, name: Optional[str] = None
    ) -> Tuple[str, str]:
        if name is not None:
            _, version = parse_cluster_compute_name_version(name)
            if version is not None:
                raise ValueError(
                    "A version tag cannot be provided when creating a compute config. "
                    "The latest version tag will be generated and returned."
                )

        """Register the provided compute config and return its internal ID."""
        compute_config_api_model = self._convert_compute_config_to_api_model(
            compute_config
        )
        full_name, compute_config_id = self.client.create_compute_config(
            compute_config_api_model, name=name
        )
        self.logger.info(f"Created compute config: '{full_name}'")
        ui_url = self.client.get_compute_config_ui_url(
            compute_config_id, cloud_id=compute_config_api_model.cloud_id
        )
        self.logger.info(f"View the compute config in the UI: '{ui_url}'")
        return full_name, compute_config_id

    def _convert_api_model_to_advanced_instance_config(
        self,
        api_model: Union[DecoratedComputeTemplate, ComputeNodeType, WorkerNodeType],
        *,
        cloud: Cloud,
    ) -> Optional[Dict]:
        if cloud.provider == CloudProviders.AWS:
            return api_model.aws_advanced_configurations_json or None
        elif cloud.provider == CloudProviders.GCP:
            return api_model.gcp_advanced_configurations_json or None
        else:
            return None

    def _convert_api_model_to_resource_dict(
        self, resources: Optional[Resources]
    ) -> Optional[Dict[str, float]]:
        # Flatten the resource dict returned by the API and strip `None` values.
        if resources is None:
            return None

        return {
            k: v
            for k, v in {
                "CPU": resources.cpu,
                "GPU": resources.gpu,
                "memory": resources.memory,
                "object_store_memory": resources.object_store_memory,
                **(resources.custom_resources or {}),
            }.items()
            if v is not None
        }

    def _convert_api_model_to_head_node_config(
        self, api_model: ComputeNodeType, *, cloud: Cloud
    ) -> HeadNodeConfig:
        return HeadNodeConfig(
            instance_type=api_model.instance_type,
            resources=self._convert_api_model_to_resource_dict(api_model.resources),
            advanced_instance_config=self._convert_api_model_to_advanced_instance_config(
                api_model, cloud=cloud,
            ),
        )

    def _convert_api_models_to_worker_node_group_configs(
        self, api_models: List[WorkerNodeType], *, cloud: Cloud
    ) -> List[WorkerNodeGroupConfig]:
        # TODO(edoakes): support advanced_instance_config.
        configs = []
        for api_model in api_models:
            if api_model.use_spot and api_model.fallback_to_ondemand:
                market_type = MarketType.PREFER_SPOT
            elif api_model.use_spot:
                market_type = MarketType.SPOT
            else:
                market_type = MarketType.ON_DEMAND

            min_nodes = api_model.min_workers
            if min_nodes is None:
                min_nodes = 0

            max_nodes = api_model.max_workers
            if max_nodes is None:
                # TODO(edoakes): this defaulting to 10 seems like really strange
                # behavior here but I'm copying what the UI does. In Shomil's new
                # API let's not make these optional.
                max_nodes = 10

            configs.append(
                WorkerNodeGroupConfig(
                    name=api_model.name,
                    instance_type=api_model.instance_type,
                    resources=self._convert_api_model_to_resource_dict(
                        api_model.resources
                    ),
                    advanced_instance_config=self._convert_api_model_to_advanced_instance_config(
                        api_model, cloud=cloud,
                    ),
                    min_nodes=min_nodes,
                    max_nodes=max_nodes,
                    market_type=market_type,
                )
            )

        return configs

    def _convert_api_model_to_compute_config_version(
        self, api_model: DecoratedComputeTemplate  # noqa: ARG002
    ) -> ComputeConfigVersion:
        api_model_config: DecoratedComputeTemplateConfig = api_model.config
        cloud = self.client.get_cloud(cloud_id=api_model_config.cloud_id)
        if cloud is None:
            raise RuntimeError(
                f"Cloud with ID '{api_model_config.cloud_id}' not found. "
                "This should never happen; please reach out to Anyscale support."
            )

        worker_nodes = None
        if not api_model_config.auto_select_worker_config:
            worker_nodes = self._convert_api_models_to_worker_node_group_configs(
                api_model_config.worker_node_types, cloud=cloud
            )

        zones = None
        # NOTE(edoakes): the API returns '["any"]' if no AZs are passed in on the creation path.
        if api_model_config.allowed_azs not in [["any"], []]:
            zones = api_model_config.allowed_azs

        max_resources = {}
        enable_cross_zone_scaling = False
        if api_model_config.flags is not None:
            enable_cross_zone_scaling = api_model_config.flags.get(
                "allow-cross-zone-autoscaling", False
            )
            max_cpus = api_model_config.flags.get("max-cpus", None)
            if max_cpus is not None:
                max_resources["CPU"] = max_cpus
            max_gpus = api_model_config.flags.get("max-gpus", None)
            if max_gpus is not None:
                max_resources["GPU"] = max_gpus

        return ComputeConfigVersion(
            name=f"{api_model.name}:{api_model.version}",
            id=api_model.id,
            config=ComputeConfig(
                cloud=cloud.name,
                zones=zones,
                advanced_instance_config=self._convert_api_model_to_advanced_instance_config(
                    api_model_config, cloud=cloud,
                ),
                enable_cross_zone_scaling=enable_cross_zone_scaling,
                head_node=self._convert_api_model_to_head_node_config(
                    api_model_config.head_node_type, cloud=cloud
                ),
                worker_nodes=worker_nodes,  # type: ignore
                max_resources=max_resources or None,
            ),
        )

    def _resolve_id(
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        include_archived: bool = False,
    ) -> str:
        if id is not None:
            compute_config_id = id
        elif name is not None:
            compute_config_id = self.client.get_compute_config_id(
                compute_config_name=name, include_archived=include_archived
            )
            if compute_config_id is None:
                raise RuntimeError(f"Compute config '{name}' not found.")
        else:
            raise ValueError("Either name or ID must be provided.")

        return compute_config_id

    def get_compute_config(
        self,
        *,
        name: Optional[str] = None,
        id: Optional[str] = None,  # noqa: A002
        include_archived: bool = False,
    ) -> ComputeConfigVersion:
        """Get the compute config for the provided name.

        The name can contain an optional version, e.g., '<name>:<version>'.
        If no version is provided, the latest one will be returned.
        """
        compute_config_id = self._resolve_id(
            name=name, id=id, include_archived=include_archived
        )
        compute_config = self.client.get_compute_config(compute_config_id)
        if compute_config is None:
            raise RuntimeError(
                f"Compute config with ID '{compute_config_id}' not found.'"
            )
        return self._convert_api_model_to_compute_config_version(compute_config)

    def archive_compute_config(
        self, *, name: Optional[str] = None, id: Optional[str] = None  # noqa: A002
    ):
        compute_config_id = self._resolve_id(name=name, id=id)
        self.client.archive_compute_config(compute_config_id=compute_config_id)
        self.logger.info("Compute config is successfully archived.")



from typing import IO

import click

import anyscale


@click.group(
    "image", help="Manage images to define dependencies on Anyscale.",
)
def image_cli() -> None:
    pass


@image_cli.command(
    name="build", help=("Build an image from a Containerfile."),
)
@click.option(
    "--containerfile",
    "-f",
    help="Path to the Containerfile.",
    type=click.File("rb"),
    required=True,
)
@click.option(
    "--name",
    "-n",
    help="Name for the image. If the image with the same name already exists, a new version will be built. Otherwise, a new image will be created.",
    required=True,
    type=str,
)
def build(containerfile: IO[bytes], name: str) -> None:
    containerfile_str = containerfile.read().decode("utf-8")
    image_uri = anyscale.image.build(containerfile_str, name=name)
    print(f"Image built successfully with URI: {image_uri}")



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DecoratedBuild(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'application_template_id': 'str',
        'config_json': 'AppConfigConfigSchema',
        'containerfile': 'str',
        'docker_image_name': 'str',
        'id': 'str',
        'application_template': 'AppConfig',
        'revision': 'int',
        'creator_id': 'str',
        'error_message': 'str',
        'status': 'BuildStatus',
        'created_at': 'datetime',
        'last_modified_at': 'datetime',
        'deleted_at': 'datetime',
        'is_byod': 'bool',
        'registry_login_secret': 'str',
        'cloud_id': 'str',
        'digest': 'str',
        'creator': 'MiniUser',
        'byod_ray_version': 'str'
    }

    attribute_map = {
        'application_template_id': 'application_template_id',
        'config_json': 'config_json',
        'containerfile': 'containerfile',
        'docker_image_name': 'docker_image_name',
        'id': 'id',
        'application_template': 'application_template',
        'revision': 'revision',
        'creator_id': 'creator_id',
        'error_message': 'error_message',
        'status': 'status',
        'created_at': 'created_at',
        'last_modified_at': 'last_modified_at',
        'deleted_at': 'deleted_at',
        'is_byod': 'is_byod',
        'registry_login_secret': 'registry_login_secret',
        'cloud_id': 'cloud_id',
        'digest': 'digest',
        'creator': 'creator',
        'byod_ray_version': 'byod_ray_version'
    }

    def __init__(self, application_template_id=None, config_json=None, containerfile=None, docker_image_name=None, id=None, application_template=None, revision=None, creator_id=None, error_message=None, status=None, created_at=None, last_modified_at=None, deleted_at=None, is_byod=None, registry_login_secret=None, cloud_id=None, digest=None, creator=None, byod_ray_version=None, local_vars_configuration=None):  # noqa: E501
        """DecoratedBuild - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._application_template_id = None
        self._config_json = None
        self._containerfile = None
        self._docker_image_name = None
        self._id = None
        self._application_template = None
        self._revision = None
        self._creator_id = None
        self._error_message = None
        self._status = None
        self._created_at = None
        self._last_modified_at = None
        self._deleted_at = None
        self._is_byod = None
        self._registry_login_secret = None
        self._cloud_id = None
        self._digest = None
        self._creator = None
        self._byod_ray_version = None
        self.discriminator = None

        self.application_template_id = application_template_id
        if config_json is not None:
            self.config_json = config_json
        if containerfile is not None:
            self.containerfile = containerfile
        if docker_image_name is not None:
            self.docker_image_name = docker_image_name
        self.id = id
        self.application_template = application_template
        self.revision = revision
        self.creator_id = creator_id
        if error_message is not None:
            self.error_message = error_message
        self.status = status
        self.created_at = created_at
        self.last_modified_at = last_modified_at
        if deleted_at is not None:
            self.deleted_at = deleted_at
        self.is_byod = is_byod
        if registry_login_secret is not None:
            self.registry_login_secret = registry_login_secret
        if cloud_id is not None:
            self.cloud_id = cloud_id
        if digest is not None:
            self.digest = digest
        self.creator = creator
        if byod_ray_version is not None:
            self.byod_ray_version = byod_ray_version

    @property
    def application_template_id(self):
        """Gets the application_template_id of this DecoratedBuild.  # noqa: E501

        ID of the App Config this Build belongs to.  # noqa: E501

        :return: The application_template_id of this DecoratedBuild.  # noqa: E501
        :rtype: str
        """
        return self._application_template_id

    @application_template_id.setter
    def application_template_id(self, application_template_id):
        """Sets the application_template_id of this DecoratedBuild.

        ID of the App Config this Build belongs to.  # noqa: E501

        :param application_template_id: The application_template_id of this DecoratedBuild.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and application_template_id is None:  # noqa: E501
            raise ValueError("Invalid value for `application_template_id`, must not be `None`")  # noqa: E501

        self._application_template_id = application_template_id

    @property
    def config_json(self):
        """Gets the config_json of this DecoratedBuild.  # noqa: E501

        Config JSON used to create this Build.  # noqa: E501

        :return: The config_json of this DecoratedBuild.  # noqa: E501
        :rtype: AppConfigConfigSchema
        """
        return self._config_json

    @config_json.setter
    def config_json(self, config_json):
        """Sets the config_json of this DecoratedBuild.

        Config JSON used to create this Build.  # noqa: E501

        :param config_json: The config_json of this DecoratedBuild.  # noqa: E501
        :type: AppConfigConfigSchema
        """

        self._config_json = config_json

    @property
    def containerfile(self):
        """Gets the containerfile of this DecoratedBuild.  # noqa: E501

        The containerfile used to build the image.  # noqa: E501

        :return: The containerfile of this DecoratedBuild.  # noqa: E501
        :rtype: str
        """
        return self._containerfile

    @containerfile.setter
    def containerfile(self, containerfile):
        """Sets the containerfile of this DecoratedBuild.

        The containerfile used to build the image.  # noqa: E501

        :param containerfile: The containerfile of this DecoratedBuild.  # noqa: E501
        :type: str
        """

        self._containerfile = containerfile

    @property
    def docker_image_name(self):
        """Gets the docker_image_name of this DecoratedBuild.  # noqa: E501

        The name of the docker image for this build.  # noqa: E501

        :return: The docker_image_name of this DecoratedBuild.  # noqa: E501
        :rtype: str
        """
        return self._docker_image_name

    @docker_image_name.setter
    def docker_image_name(self, docker_image_name):
        """Sets the docker_image_name of this DecoratedBuild.

        The name of the docker image for this build.  # noqa: E501

        :param docker_image_name: The docker_image_name of this DecoratedBuild.  # noqa: E501
        :type: str
        """

        self._docker_image_name = docker_image_name

    @property
    def id(self):
        """Gets the id of this DecoratedBuild.  # noqa: E501

        Server assigned unique identifier.  # noqa: E501

        :return: The id of this DecoratedBuild.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this DecoratedBuild.

        Server assigned unique identifier.  # noqa: E501

        :param id: The id of this DecoratedBuild.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def application_template(self):
        """Gets the application_template of this DecoratedBuild.  # noqa: E501

        App Config this Build belongs to.  # noqa: E501

        :return: The application_template of this DecoratedBuild.  # noqa: E501
        :rtype: AppConfig
        """
        return self._application_template

    @application_template.setter
    def application_template(self, application_template):
        """Sets the application_template of this DecoratedBuild.

        App Config this Build belongs to.  # noqa: E501

        :param application_template: The application_template of this DecoratedBuild.  # noqa: E501
        :type: AppConfig
        """
        if self.local_vars_configuration.client_side_validation and application_template is None:  # noqa: E501
            raise ValueError("Invalid value for `application_template`, must not be `None`")  # noqa: E501

        self._application_template = application_template

    @property
    def revision(self):
        """Gets the revision of this DecoratedBuild.  # noqa: E501

        Auto incrementing version number for this Build  # noqa: E501

        :return: The revision of this DecoratedBuild.  # noqa: E501
        :rtype: int
        """
        return self._revision

    @revision.setter
    def revision(self, revision):
        """Sets the revision of this DecoratedBuild.

        Auto incrementing version number for this Build  # noqa: E501

        :param revision: The revision of this DecoratedBuild.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and revision is None:  # noqa: E501
            raise ValueError("Invalid value for `revision`, must not be `None`")  # noqa: E501

        self._revision = revision

    @property
    def creator_id(self):
        """Gets the creator_id of this DecoratedBuild.  # noqa: E501

        ID of the user who created this Build.  # noqa: E501

        :return: The creator_id of this DecoratedBuild.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this DecoratedBuild.

        ID of the user who created this Build.  # noqa: E501

        :param creator_id: The creator_id of this DecoratedBuild.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def error_message(self):
        """Gets the error_message of this DecoratedBuild.  # noqa: E501

        Detailed error message. This will only be populated if the Build operation failed.  # noqa: E501

        :return: The error_message of this DecoratedBuild.  # noqa: E501
        :rtype: str
        """
        return self._error_message

    @error_message.setter
    def error_message(self, error_message):
        """Sets the error_message of this DecoratedBuild.

        Detailed error message. This will only be populated if the Build operation failed.  # noqa: E501

        :param error_message: The error_message of this DecoratedBuild.  # noqa: E501
        :type: str
        """

        self._error_message = error_message

    @property
    def status(self):
        """Gets the status of this DecoratedBuild.  # noqa: E501

             Status of the Build.      `pending` - Build operation is queued and has not started yet.     `in_progress` - Build operation is in progress.     `succeeded` - Build operation completed successfully.     `failed` - Build operation completed unsuccessfully.     `pending_cancellation` - Build operation is marked for cancellation.     `cancelled` - Build operation was cancelled before it completed.       # noqa: E501

        :return: The status of this DecoratedBuild.  # noqa: E501
        :rtype: BuildStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this DecoratedBuild.

             Status of the Build.      `pending` - Build operation is queued and has not started yet.     `in_progress` - Build operation is in progress.     `succeeded` - Build operation completed successfully.     `failed` - Build operation completed unsuccessfully.     `pending_cancellation` - Build operation is marked for cancellation.     `cancelled` - Build operation was cancelled before it completed.       # noqa: E501

        :param status: The status of this DecoratedBuild.  # noqa: E501
        :type: BuildStatus
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def created_at(self):
        """Gets the created_at of this DecoratedBuild.  # noqa: E501

        Timestamp of when this Build was created.  # noqa: E501

        :return: The created_at of this DecoratedBuild.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this DecoratedBuild.

        Timestamp of when this Build was created.  # noqa: E501

        :param created_at: The created_at of this DecoratedBuild.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def last_modified_at(self):
        """Gets the last_modified_at of this DecoratedBuild.  # noqa: E501

        Timestamp of when this Build was last updated.  # noqa: E501

        :return: The last_modified_at of this DecoratedBuild.  # noqa: E501
        :rtype: datetime
        """
        return self._last_modified_at

    @last_modified_at.setter
    def last_modified_at(self, last_modified_at):
        """Sets the last_modified_at of this DecoratedBuild.

        Timestamp of when this Build was last updated.  # noqa: E501

        :param last_modified_at: The last_modified_at of this DecoratedBuild.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and last_modified_at is None:  # noqa: E501
            raise ValueError("Invalid value for `last_modified_at`, must not be `None`")  # noqa: E501

        self._last_modified_at = last_modified_at

    @property
    def deleted_at(self):
        """Gets the deleted_at of this DecoratedBuild.  # noqa: E501

        Timestamp of when this Build was deleted.  # noqa: E501

        :return: The deleted_at of this DecoratedBuild.  # noqa: E501
        :rtype: datetime
        """
        return self._deleted_at

    @deleted_at.setter
    def deleted_at(self, deleted_at):
        """Sets the deleted_at of this DecoratedBuild.

        Timestamp of when this Build was deleted.  # noqa: E501

        :param deleted_at: The deleted_at of this DecoratedBuild.  # noqa: E501
        :type: datetime
        """

        self._deleted_at = deleted_at

    @property
    def is_byod(self):
        """Gets the is_byod of this DecoratedBuild.  # noqa: E501

        True if the image URI used in this build was user-specified.  # noqa: E501

        :return: The is_byod of this DecoratedBuild.  # noqa: E501
        :rtype: bool
        """
        return self._is_byod

    @is_byod.setter
    def is_byod(self, is_byod):
        """Sets the is_byod of this DecoratedBuild.

        True if the image URI used in this build was user-specified.  # noqa: E501

        :param is_byod: The is_byod of this DecoratedBuild.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and is_byod is None:  # noqa: E501
            raise ValueError("Invalid value for `is_byod`, must not be `None`")  # noqa: E501

        self._is_byod = is_byod

    @property
    def registry_login_secret(self):
        """Gets the registry_login_secret of this DecoratedBuild.  # noqa: E501

        The name or identifier of a secret containing credentials to authenticate to the docker registry hosting the image.  # noqa: E501

        :return: The registry_login_secret of this DecoratedBuild.  # noqa: E501
        :rtype: str
        """
        return self._registry_login_secret

    @registry_login_secret.setter
    def registry_login_secret(self, registry_login_secret):
        """Sets the registry_login_secret of this DecoratedBuild.

        The name or identifier of a secret containing credentials to authenticate to the docker registry hosting the image.  # noqa: E501

        :param registry_login_secret: The registry_login_secret of this DecoratedBuild.  # noqa: E501
        :type: str
        """

        self._registry_login_secret = registry_login_secret

    @property
    def cloud_id(self):
        """Gets the cloud_id of this DecoratedBuild.  # noqa: E501

        The build cloud associated with this build. If None, the build is a v1 build.  # noqa: E501

        :return: The cloud_id of this DecoratedBuild.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this DecoratedBuild.

        The build cloud associated with this build. If None, the build is a v1 build.  # noqa: E501

        :param cloud_id: The cloud_id of this DecoratedBuild.  # noqa: E501
        :type: str
        """

        self._cloud_id = cloud_id

    @property
    def digest(self):
        """Gets the digest of this DecoratedBuild.  # noqa: E501

        The digest of the image created by this build.  # noqa: E501

        :return: The digest of this DecoratedBuild.  # noqa: E501
        :rtype: str
        """
        return self._digest

    @digest.setter
    def digest(self, digest):
        """Sets the digest of this DecoratedBuild.

        The digest of the image created by this build.  # noqa: E501

        :param digest: The digest of this DecoratedBuild.  # noqa: E501
        :type: str
        """

        self._digest = digest

    @property
    def creator(self):
        """Gets the creator of this DecoratedBuild.  # noqa: E501

        Read-only field. Information about the creator.  # noqa: E501

        :return: The creator of this DecoratedBuild.  # noqa: E501
        :rtype: MiniUser
        """
        return self._creator

    @creator.setter
    def creator(self, creator):
        """Sets the creator of this DecoratedBuild.

        Read-only field. Information about the creator.  # noqa: E501

        :param creator: The creator of this DecoratedBuild.  # noqa: E501
        :type: MiniUser
        """
        if self.local_vars_configuration.client_side_validation and creator is None:  # noqa: E501
            raise ValueError("Invalid value for `creator`, must not be `None`")  # noqa: E501

        self._creator = creator

    @property
    def byod_ray_version(self):
        """Gets the byod_ray_version of this DecoratedBuild.  # noqa: E501

        Read-only field. If this is a BYOD build, contains the user specified Ray version.  # noqa: E501

        :return: The byod_ray_version of this DecoratedBuild.  # noqa: E501
        :rtype: str
        """
        return self._byod_ray_version

    @byod_ray_version.setter
    def byod_ray_version(self, byod_ray_version):
        """Sets the byod_ray_version of this DecoratedBuild.

        Read-only field. If this is a BYOD build, contains the user specified Ray version.  # noqa: E501

        :param byod_ray_version: The byod_ray_version of this DecoratedBuild.  # noqa: E501
        :type: str
        """

        self._byod_ray_version = byod_ray_version

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DecoratedBuild):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DecoratedBuild):
            return True

        return self.to_dict() != other.to_dict()



import click

from anyscale.cli_logger import BlockLogger
from anyscale.controllers.service_account_controller import ServiceAccountController
from anyscale.util import validate_non_negative_arg


log = BlockLogger()  # CLI Logger


@click.group(
    "service-account",
    short_help="Manage service accounts for your anyscale workloads.",
)
def service_account_cli() -> None:
    pass


@service_account_cli.command(name="create", help="Create a service account.")
@click.option("--name", "-n", help="Name for the service account.", type=str)
def create(name: str) -> None:
    ServiceAccountController().create_service_account(name)


@service_account_cli.command(
    name="create-api-key", help="Create a new API key for a service account."
)
@click.option(
    "--email", help="Email of the service account to create the new key for.", type=str
)
def create_api_key(email: str) -> None:
    ServiceAccountController().create_new_service_account_api_key(email)


@service_account_cli.command(name="list", help="List all service accounts.")
@click.option(
    "--max-items",
    required=False,
    default=20,
    type=int,
    help="Max items to show in list.",
    callback=validate_non_negative_arg,
)
def list_service_accounts(max_items: int) -> None:
    ServiceAccountController().list_service_accounts(max_items)


@service_account_cli.command(name="delete", help="Delete a service account.")
@click.option("--email", help="Email of the service account to delete.", type=str)
def delete(email) -> None:
    ServiceAccountController().delete_service_account(email)


@service_account_cli.command(
    name="rotate-api-keys", help="Rotate all api keys of a service account."
)
@click.option(
    "--email", help="Rotate API keys for the service account with this email.", type=str
)
def rotate_api_keys(email) -> None:
    ServiceAccountController().rotate_service_account_api_keys(email)



from anyscale.compute_config.commands import archive, create, get
from anyscale.compute_config.models import (
    ComputeConfig,
    ComputeConfigVersion,
    HeadNodeConfig,
    WorkerNodeGroupConfig,
)



from anyscale._private.sdk import sdk_command
from anyscale.image._private.image_sdk import ImageSDK


_IMAGE_SDK_SINGLETON_KEY = "image_sdk"

_BUILD_FROM_CONTAINERFILE_EXAMPLE = """
import anyscale

containerfile = '''
FROM anyscale/ray:2.21.0-py39
RUN pip install --no-cache-dir pandas
'''

image_uri: str = anyscale.image.build(containerfile, name="mycoolimage")
"""


@sdk_command(
    _IMAGE_SDK_SINGLETON_KEY,
    ImageSDK,
    doc_py_example=_BUILD_FROM_CONTAINERFILE_EXAMPLE,
    arg_docstrings={
        "name": "The name of the image.",
        "containerfile": "The content of the Containerfile.",
    },
)
def build(containerfile: str, *, name: str, _sdk: ImageSDK) -> str:
    build_id = _sdk.build_image_from_containerfile(name, containerfile)
    image_uri = _sdk.get_image_uri_from_build_id(build_id)
    if image_uri:
        return image_uri.image_uri
    raise RuntimeError(
        f"This is a bug! Failed to get image uri for build {build_id} that just created."
    )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class AicaEndpoint(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'str',
        'aviary_version': 'str',
        'service_id': 'str',
        'model_config': 'object',
        'id': 'str',
        'created_at': 'datetime',
        'creator_id': 'str',
        'creator_email': 'str',
        'creator_deleted_at': 'datetime',
        'archived_at': 'datetime',
        'cloud_id': 'str',
        'project_id': 'str',
        'organization_id': 'str',
        'aica_observability_urls': 'AicaObservabilityUrls',
        'state': 'ServiceEventCurrentState'
    }

    attribute_map = {
        'name': 'name',
        'aviary_version': 'aviary_version',
        'service_id': 'service_id',
        'model_config': 'model_config',
        'id': 'id',
        'created_at': 'created_at',
        'creator_id': 'creator_id',
        'creator_email': 'creator_email',
        'creator_deleted_at': 'creator_deleted_at',
        'archived_at': 'archived_at',
        'cloud_id': 'cloud_id',
        'project_id': 'project_id',
        'organization_id': 'organization_id',
        'aica_observability_urls': 'aica_observability_urls',
        'state': 'state'
    }

    def __init__(self, name=None, aviary_version=None, service_id=None, model_config=None, id=None, created_at=None, creator_id=None, creator_email=None, creator_deleted_at=None, archived_at=None, cloud_id=None, project_id=None, organization_id=None, aica_observability_urls=None, state=None, local_vars_configuration=None):  # noqa: E501
        """AicaEndpoint - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._aviary_version = None
        self._service_id = None
        self._model_config = None
        self._id = None
        self._created_at = None
        self._creator_id = None
        self._creator_email = None
        self._creator_deleted_at = None
        self._archived_at = None
        self._cloud_id = None
        self._project_id = None
        self._organization_id = None
        self._aica_observability_urls = None
        self._state = None
        self.discriminator = None

        self.name = name
        self.aviary_version = aviary_version
        self.service_id = service_id
        self.model_config = model_config
        self.id = id
        self.created_at = created_at
        self.creator_id = creator_id
        self.creator_email = creator_email
        if creator_deleted_at is not None:
            self.creator_deleted_at = creator_deleted_at
        if archived_at is not None:
            self.archived_at = archived_at
        self.cloud_id = cloud_id
        self.project_id = project_id
        self.organization_id = organization_id
        self.aica_observability_urls = aica_observability_urls
        self.state = state

    @property
    def name(self):
        """Gets the name of this AicaEndpoint.  # noqa: E501

        Name of the endpoint to be created.  # noqa: E501

        :return: The name of this AicaEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this AicaEndpoint.

        Name of the endpoint to be created.  # noqa: E501

        :param name: The name of this AicaEndpoint.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def aviary_version(self):
        """Gets the aviary_version of this AicaEndpoint.  # noqa: E501

        The version of aviary that this endpoint is running.  # noqa: E501

        :return: The aviary_version of this AicaEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._aviary_version

    @aviary_version.setter
    def aviary_version(self, aviary_version):
        """Sets the aviary_version of this AicaEndpoint.

        The version of aviary that this endpoint is running.  # noqa: E501

        :param aviary_version: The aviary_version of this AicaEndpoint.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and aviary_version is None:  # noqa: E501
            raise ValueError("Invalid value for `aviary_version`, must not be `None`")  # noqa: E501

        self._aviary_version = aviary_version

    @property
    def service_id(self):
        """Gets the service_id of this AicaEndpoint.  # noqa: E501

        Id of the associated service.  # noqa: E501

        :return: The service_id of this AicaEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._service_id

    @service_id.setter
    def service_id(self, service_id):
        """Sets the service_id of this AicaEndpoint.

        Id of the associated service.  # noqa: E501

        :param service_id: The service_id of this AicaEndpoint.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and service_id is None:  # noqa: E501
            raise ValueError("Invalid value for `service_id`, must not be `None`")  # noqa: E501

        self._service_id = service_id

    @property
    def model_config(self):
        """Gets the model_config of this AicaEndpoint.  # noqa: E501

        The configuration for the endpoints models. Key is the model id.  # noqa: E501

        :return: The model_config of this AicaEndpoint.  # noqa: E501
        :rtype: object
        """
        return self._model_config

    @model_config.setter
    def model_config(self, model_config):
        """Sets the model_config of this AicaEndpoint.

        The configuration for the endpoints models. Key is the model id.  # noqa: E501

        :param model_config: The model_config of this AicaEndpoint.  # noqa: E501
        :type: object
        """
        if self.local_vars_configuration.client_side_validation and model_config is None:  # noqa: E501
            raise ValueError("Invalid value for `model_config`, must not be `None`")  # noqa: E501

        self._model_config = model_config

    @property
    def id(self):
        """Gets the id of this AicaEndpoint.  # noqa: E501

        Server assigned unique identifier of the endpoint.  # noqa: E501

        :return: The id of this AicaEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this AicaEndpoint.

        Server assigned unique identifier of the endpoint.  # noqa: E501

        :param id: The id of this AicaEndpoint.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def created_at(self):
        """Gets the created_at of this AicaEndpoint.  # noqa: E501

        Time at which endpoint was created.  # noqa: E501

        :return: The created_at of this AicaEndpoint.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this AicaEndpoint.

        Time at which endpoint was created.  # noqa: E501

        :param created_at: The created_at of this AicaEndpoint.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def creator_id(self):
        """Gets the creator_id of this AicaEndpoint.  # noqa: E501

        Identifier of user who created the endpoint.  # noqa: E501

        :return: The creator_id of this AicaEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this AicaEndpoint.

        Identifier of user who created the endpoint.  # noqa: E501

        :param creator_id: The creator_id of this AicaEndpoint.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def creator_email(self):
        """Gets the creator_email of this AicaEndpoint.  # noqa: E501

        Email of user who created the endpoint.  # noqa: E501

        :return: The creator_email of this AicaEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._creator_email

    @creator_email.setter
    def creator_email(self, creator_email):
        """Sets the creator_email of this AicaEndpoint.

        Email of user who created the endpoint.  # noqa: E501

        :param creator_email: The creator_email of this AicaEndpoint.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_email is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_email`, must not be `None`")  # noqa: E501

        self._creator_email = creator_email

    @property
    def creator_deleted_at(self):
        """Gets the creator_deleted_at of this AicaEndpoint.  # noqa: E501

        Timestamp of when the user who created the endpoint was deleted.  # noqa: E501

        :return: The creator_deleted_at of this AicaEndpoint.  # noqa: E501
        :rtype: datetime
        """
        return self._creator_deleted_at

    @creator_deleted_at.setter
    def creator_deleted_at(self, creator_deleted_at):
        """Sets the creator_deleted_at of this AicaEndpoint.

        Timestamp of when the user who created the endpoint was deleted.  # noqa: E501

        :param creator_deleted_at: The creator_deleted_at of this AicaEndpoint.  # noqa: E501
        :type: datetime
        """

        self._creator_deleted_at = creator_deleted_at

    @property
    def archived_at(self):
        """Gets the archived_at of this AicaEndpoint.  # noqa: E501

        The time in which this instance is archived.  # noqa: E501

        :return: The archived_at of this AicaEndpoint.  # noqa: E501
        :rtype: datetime
        """
        return self._archived_at

    @archived_at.setter
    def archived_at(self, archived_at):
        """Sets the archived_at of this AicaEndpoint.

        The time in which this instance is archived.  # noqa: E501

        :param archived_at: The archived_at of this AicaEndpoint.  # noqa: E501
        :type: datetime
        """

        self._archived_at = archived_at

    @property
    def cloud_id(self):
        """Gets the cloud_id of this AicaEndpoint.  # noqa: E501

        The cloud id for the endpoint.  # noqa: E501

        :return: The cloud_id of this AicaEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this AicaEndpoint.

        The cloud id for the endpoint.  # noqa: E501

        :param cloud_id: The cloud_id of this AicaEndpoint.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def project_id(self):
        """Gets the project_id of this AicaEndpoint.  # noqa: E501

        Id of the project that this endpoint belongs to.  # noqa: E501

        :return: The project_id of this AicaEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this AicaEndpoint.

        Id of the project that this endpoint belongs to.  # noqa: E501

        :param project_id: The project_id of this AicaEndpoint.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_id is None:  # noqa: E501
            raise ValueError("Invalid value for `project_id`, must not be `None`")  # noqa: E501

        self._project_id = project_id

    @property
    def organization_id(self):
        """Gets the organization_id of this AicaEndpoint.  # noqa: E501

        Id of the organization that this endpoint belongs to  # noqa: E501

        :return: The organization_id of this AicaEndpoint.  # noqa: E501
        :rtype: str
        """
        return self._organization_id

    @organization_id.setter
    def organization_id(self, organization_id):
        """Sets the organization_id of this AicaEndpoint.

        Id of the organization that this endpoint belongs to  # noqa: E501

        :param organization_id: The organization_id of this AicaEndpoint.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and organization_id is None:  # noqa: E501
            raise ValueError("Invalid value for `organization_id`, must not be `None`")  # noqa: E501

        self._organization_id = organization_id

    @property
    def aica_observability_urls(self):
        """Gets the aica_observability_urls of this AicaEndpoint.  # noqa: E501

        A JSON object with useful urls pointing to Grafana dashboards.  # noqa: E501

        :return: The aica_observability_urls of this AicaEndpoint.  # noqa: E501
        :rtype: AicaObservabilityUrls
        """
        return self._aica_observability_urls

    @aica_observability_urls.setter
    def aica_observability_urls(self, aica_observability_urls):
        """Sets the aica_observability_urls of this AicaEndpoint.

        A JSON object with useful urls pointing to Grafana dashboards.  # noqa: E501

        :param aica_observability_urls: The aica_observability_urls of this AicaEndpoint.  # noqa: E501
        :type: AicaObservabilityUrls
        """
        if self.local_vars_configuration.client_side_validation and aica_observability_urls is None:  # noqa: E501
            raise ValueError("Invalid value for `aica_observability_urls`, must not be `None`")  # noqa: E501

        self._aica_observability_urls = aica_observability_urls

    @property
    def state(self):
        """Gets the state of this AicaEndpoint.  # noqa: E501

        The status of the endpoint.  # noqa: E501

        :return: The state of this AicaEndpoint.  # noqa: E501
        :rtype: ServiceEventCurrentState
        """
        return self._state

    @state.setter
    def state(self, state):
        """Sets the state of this AicaEndpoint.

        The status of the endpoint.  # noqa: E501

        :param state: The state of this AicaEndpoint.  # noqa: E501
        :type: ServiceEventCurrentState
        """
        if self.local_vars_configuration.client_side_validation and state is None:  # noqa: E501
            raise ValueError("Invalid value for `state`, must not be `None`")  # noqa: E501

        self._state = state

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, AicaEndpoint):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, AicaEndpoint):
            return True

        return self.to_dict() != other.to_dict()



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
import asyncio

import pytest

from shared_anyscale_utils.utils.asyncio import chain_future, create_future


@pytest.mark.parametrize("result", [42, Exception("failed!")])
async def test_chain_future_completes(result):
    one = create_future()
    other = create_future()

    chain_future(one, other)

    if isinstance(result, Exception):
        one.set_exception(result)
    else:
        one.set_result(result)

    await asyncio.gather(other, return_exceptions=True)

    if isinstance(result, Exception):
        assert str(other.exception()) == str(result)
    else:
        assert other.result() == 42


async def test_chain_future_cancellation():
    one = create_future()
    other = create_future()

    chain_future(one, other)

    one.cancel()

    # NOTE: We're using `gather` to retrieve `CancelledError` since `future.exception` will
    #       raise it instead of returning it
    e = (await asyncio.gather(other, return_exceptions=True))[0]

    assert type(e) == asyncio.CancelledError



from typing import Optional, Union

from anyscale._private.sdk import sdk_command
from anyscale.job._private.job_sdk import JobSDK
from anyscale.job.models import JobConfig, JobState


_JOB_SDK_SINGLETON_KEY = "job_sdk"

_SUBMIT_EXAMPLE = """
import anyscale
from anyscale.job.models import JobConfig

anyscale.job.submit(
    JobConfig(
        name="my-job",
        entrypoint="python main.py",
        working_dir=".",
    ),
)
"""


@sdk_command(
    _JOB_SDK_SINGLETON_KEY,
    JobSDK,
    doc_py_example=_SUBMIT_EXAMPLE,
    arg_docstrings={"config": "The config options defining the job.",},
)
def submit(config: JobConfig, *, _sdk: JobSDK) -> str:
    """Submit a job.

    Returns the id of the submitted job.
    """
    return _sdk.submit(config)


_STATUS_EXAMPLE = """
import anyscale
from anyscale.job.models import JobStatus

status: JobStatus = anyscale.job.status(name="my-job")
"""


@sdk_command(
    _JOB_SDK_SINGLETON_KEY,
    JobSDK,
    doc_py_example=_STATUS_EXAMPLE,
    arg_docstrings={"name": "Name of the job.", "job_id": "Unique ID of the job"},
)
def status(
    name: Optional[str] = None, job_id: Optional[str] = None, *, _sdk: JobSDK
) -> str:
    """Get the status of a job."""
    return _sdk.status(name=name, job_id=job_id)


_TERMINATE_EXAMPLE = """
import anyscale

anyscale.job.terminate(name="my-job")
"""


@sdk_command(
    _JOB_SDK_SINGLETON_KEY,
    JobSDK,
    doc_py_example=_TERMINATE_EXAMPLE,
    arg_docstrings={"name": "Name of the job.", "job_id": "Unique ID of the job"},
)
def terminate(
    name: Optional[str] = None, job_id: Optional[str] = None, *, _sdk: JobSDK
) -> str:
    """Terminate a job.

    This command is asynchronous, so it always returns immediately.

    Returns the id of the terminated job.
    """
    return _sdk.terminate(name=name, job_id=job_id)


_ARCHIVE_EXAMPLE = """
import anyscale

anyscale.job.archive(name="my-job")
"""


@sdk_command(
    _JOB_SDK_SINGLETON_KEY,
    JobSDK,
    doc_py_example=_ARCHIVE_EXAMPLE,
    arg_docstrings={"name": "Name of the job.", "job_id": "Unique ID of the job"},
)
def archive(
    name: Optional[str] = None, job_id: Optional[str] = None, *, _sdk: JobSDK
) -> str:
    """Archive a job.

    This command is asynchronous, so it always returns immediately.

    Returns the id of the archived job.
    """
    return _sdk.archive(name=name, job_id=job_id)


_WAIT_EXAMPLE = """\
import anyscale

anyscale.job.wait(name="my-job", timeout_s=180)"""


@sdk_command(
    _JOB_SDK_SINGLETON_KEY,
    JobSDK,
    doc_py_example=_WAIT_EXAMPLE,
    arg_docstrings={
        "name": "Name of the job.",
        "job_id": "Unique ID of the job",
        "state": "Target state of the job",
        "timeout_s": "Number of seconds to wait before timing out, this timeout will not affect job execution",
    },
)
def wait(
    *,
    name: Optional[str] = None,
    job_id: Optional[str] = None,
    state: Union[JobState, str] = JobState.SUCCEEDED,
    timeout_s: float = 1800,
    _sdk: JobSDK
):
    """"Wait for a job to enter a specific state."""
    _sdk.wait(name=name, job_id=job_id, state=state, timeout_s=timeout_s)



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class ClusterServicesUrls(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'webterminal_auth_url': 'str',
        'metrics_dashboard_url': 'str',
        'persistent_metrics_url': 'str',
        'connect_url': 'str',
        'jupyter_notebook_url': 'str',
        'ray_dashboard_url': 'str',
        'service_proxy_url': 'str',
        'user_service_url': 'str'
    }

    attribute_map = {
        'webterminal_auth_url': 'webterminal_auth_url',
        'metrics_dashboard_url': 'metrics_dashboard_url',
        'persistent_metrics_url': 'persistent_metrics_url',
        'connect_url': 'connect_url',
        'jupyter_notebook_url': 'jupyter_notebook_url',
        'ray_dashboard_url': 'ray_dashboard_url',
        'service_proxy_url': 'service_proxy_url',
        'user_service_url': 'user_service_url'
    }

    def __init__(self, webterminal_auth_url=None, metrics_dashboard_url=None, persistent_metrics_url=None, connect_url=None, jupyter_notebook_url=None, ray_dashboard_url=None, service_proxy_url=None, user_service_url=None, local_vars_configuration=None):  # noqa: E501
        """ClusterServicesUrls - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._webterminal_auth_url = None
        self._metrics_dashboard_url = None
        self._persistent_metrics_url = None
        self._connect_url = None
        self._jupyter_notebook_url = None
        self._ray_dashboard_url = None
        self._service_proxy_url = None
        self._user_service_url = None
        self.discriminator = None

        if webterminal_auth_url is not None:
            self.webterminal_auth_url = webterminal_auth_url
        if metrics_dashboard_url is not None:
            self.metrics_dashboard_url = metrics_dashboard_url
        if persistent_metrics_url is not None:
            self.persistent_metrics_url = persistent_metrics_url
        if connect_url is not None:
            self.connect_url = connect_url
        if jupyter_notebook_url is not None:
            self.jupyter_notebook_url = jupyter_notebook_url
        if ray_dashboard_url is not None:
            self.ray_dashboard_url = ray_dashboard_url
        if service_proxy_url is not None:
            self.service_proxy_url = service_proxy_url
        if user_service_url is not None:
            self.user_service_url = user_service_url

    @property
    def webterminal_auth_url(self):
        """Gets the webterminal_auth_url of this ClusterServicesUrls.  # noqa: E501

        URL to authenticate with the webterminal  # noqa: E501

        :return: The webterminal_auth_url of this ClusterServicesUrls.  # noqa: E501
        :rtype: str
        """
        return self._webterminal_auth_url

    @webterminal_auth_url.setter
    def webterminal_auth_url(self, webterminal_auth_url):
        """Sets the webterminal_auth_url of this ClusterServicesUrls.

        URL to authenticate with the webterminal  # noqa: E501

        :param webterminal_auth_url: The webterminal_auth_url of this ClusterServicesUrls.  # noqa: E501
        :type: str
        """

        self._webterminal_auth_url = webterminal_auth_url

    @property
    def metrics_dashboard_url(self):
        """Gets the metrics_dashboard_url of this ClusterServicesUrls.  # noqa: E501

        URL for Grafana (metrics) dashboard in the running cluster state.  # noqa: E501

        :return: The metrics_dashboard_url of this ClusterServicesUrls.  # noqa: E501
        :rtype: str
        """
        return self._metrics_dashboard_url

    @metrics_dashboard_url.setter
    def metrics_dashboard_url(self, metrics_dashboard_url):
        """Sets the metrics_dashboard_url of this ClusterServicesUrls.

        URL for Grafana (metrics) dashboard in the running cluster state.  # noqa: E501

        :param metrics_dashboard_url: The metrics_dashboard_url of this ClusterServicesUrls.  # noqa: E501
        :type: str
        """

        self._metrics_dashboard_url = metrics_dashboard_url

    @property
    def persistent_metrics_url(self):
        """Gets the persistent_metrics_url of this ClusterServicesUrls.  # noqa: E501

        URL for the persistent Grafana (metrics) dashboard in the non-running cluster state.  # noqa: E501

        :return: The persistent_metrics_url of this ClusterServicesUrls.  # noqa: E501
        :rtype: str
        """
        return self._persistent_metrics_url

    @persistent_metrics_url.setter
    def persistent_metrics_url(self, persistent_metrics_url):
        """Sets the persistent_metrics_url of this ClusterServicesUrls.

        URL for the persistent Grafana (metrics) dashboard in the non-running cluster state.  # noqa: E501

        :param persistent_metrics_url: The persistent_metrics_url of this ClusterServicesUrls.  # noqa: E501
        :type: str
        """

        self._persistent_metrics_url = persistent_metrics_url

    @property
    def connect_url(self):
        """Gets the connect_url of this ClusterServicesUrls.  # noqa: E501

        URL for Anyscale connect.  # noqa: E501

        :return: The connect_url of this ClusterServicesUrls.  # noqa: E501
        :rtype: str
        """
        return self._connect_url

    @connect_url.setter
    def connect_url(self, connect_url):
        """Sets the connect_url of this ClusterServicesUrls.

        URL for Anyscale connect.  # noqa: E501

        :param connect_url: The connect_url of this ClusterServicesUrls.  # noqa: E501
        :type: str
        """

        self._connect_url = connect_url

    @property
    def jupyter_notebook_url(self):
        """Gets the jupyter_notebook_url of this ClusterServicesUrls.  # noqa: E501

        URL for Jupyter Lab.  # noqa: E501

        :return: The jupyter_notebook_url of this ClusterServicesUrls.  # noqa: E501
        :rtype: str
        """
        return self._jupyter_notebook_url

    @jupyter_notebook_url.setter
    def jupyter_notebook_url(self, jupyter_notebook_url):
        """Sets the jupyter_notebook_url of this ClusterServicesUrls.

        URL for Jupyter Lab.  # noqa: E501

        :param jupyter_notebook_url: The jupyter_notebook_url of this ClusterServicesUrls.  # noqa: E501
        :type: str
        """

        self._jupyter_notebook_url = jupyter_notebook_url

    @property
    def ray_dashboard_url(self):
        """Gets the ray_dashboard_url of this ClusterServicesUrls.  # noqa: E501

        URL for Ray dashboard.  # noqa: E501

        :return: The ray_dashboard_url of this ClusterServicesUrls.  # noqa: E501
        :rtype: str
        """
        return self._ray_dashboard_url

    @ray_dashboard_url.setter
    def ray_dashboard_url(self, ray_dashboard_url):
        """Sets the ray_dashboard_url of this ClusterServicesUrls.

        URL for Ray dashboard.  # noqa: E501

        :param ray_dashboard_url: The ray_dashboard_url of this ClusterServicesUrls.  # noqa: E501
        :type: str
        """

        self._ray_dashboard_url = ray_dashboard_url

    @property
    def service_proxy_url(self):
        """Gets the service_proxy_url of this ClusterServicesUrls.  # noqa: E501

        URL for web services proxy (e.g. jupyter, tensorboard, etc).  # noqa: E501

        :return: The service_proxy_url of this ClusterServicesUrls.  # noqa: E501
        :rtype: str
        """
        return self._service_proxy_url

    @service_proxy_url.setter
    def service_proxy_url(self, service_proxy_url):
        """Sets the service_proxy_url of this ClusterServicesUrls.

        URL for web services proxy (e.g. jupyter, tensorboard, etc).  # noqa: E501

        :param service_proxy_url: The service_proxy_url of this ClusterServicesUrls.  # noqa: E501
        :type: str
        """

        self._service_proxy_url = service_proxy_url

    @property
    def user_service_url(self):
        """Gets the user_service_url of this ClusterServicesUrls.  # noqa: E501

        URL to access user services (e.g. Ray Serve)  # noqa: E501

        :return: The user_service_url of this ClusterServicesUrls.  # noqa: E501
        :rtype: str
        """
        return self._user_service_url

    @user_service_url.setter
    def user_service_url(self, user_service_url):
        """Sets the user_service_url of this ClusterServicesUrls.

        URL to access user services (e.g. Ray Serve)  # noqa: E501

        :param user_service_url: The user_service_url of this ClusterServicesUrls.  # noqa: E501
        :type: str
        """

        self._user_service_url = user_service_url

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ClusterServicesUrls):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ClusterServicesUrls):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DecoratedSupportRequest(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'requester_id': 'str',
        'organization_id': 'str',
        'requested_at': 'datetime',
        'expires_at': 'datetime',
        'requester': 'MiniUser',
        'organization': 'MiniOrganization'
    }

    attribute_map = {
        'id': 'id',
        'requester_id': 'requester_id',
        'organization_id': 'organization_id',
        'requested_at': 'requested_at',
        'expires_at': 'expires_at',
        'requester': 'requester',
        'organization': 'organization'
    }

    def __init__(self, id=None, requester_id=None, organization_id=None, requested_at=None, expires_at=None, requester=None, organization=None, local_vars_configuration=None):  # noqa: E501
        """DecoratedSupportRequest - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._requester_id = None
        self._organization_id = None
        self._requested_at = None
        self._expires_at = None
        self._requester = None
        self._organization = None
        self.discriminator = None

        self.id = id
        self.requester_id = requester_id
        self.organization_id = organization_id
        self.requested_at = requested_at
        self.expires_at = expires_at
        self.requester = requester
        self.organization = organization

    @property
    def id(self):
        """Gets the id of this DecoratedSupportRequest.  # noqa: E501


        :return: The id of this DecoratedSupportRequest.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this DecoratedSupportRequest.


        :param id: The id of this DecoratedSupportRequest.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def requester_id(self):
        """Gets the requester_id of this DecoratedSupportRequest.  # noqa: E501


        :return: The requester_id of this DecoratedSupportRequest.  # noqa: E501
        :rtype: str
        """
        return self._requester_id

    @requester_id.setter
    def requester_id(self, requester_id):
        """Sets the requester_id of this DecoratedSupportRequest.


        :param requester_id: The requester_id of this DecoratedSupportRequest.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and requester_id is None:  # noqa: E501
            raise ValueError("Invalid value for `requester_id`, must not be `None`")  # noqa: E501

        self._requester_id = requester_id

    @property
    def organization_id(self):
        """Gets the organization_id of this DecoratedSupportRequest.  # noqa: E501


        :return: The organization_id of this DecoratedSupportRequest.  # noqa: E501
        :rtype: str
        """
        return self._organization_id

    @organization_id.setter
    def organization_id(self, organization_id):
        """Sets the organization_id of this DecoratedSupportRequest.


        :param organization_id: The organization_id of this DecoratedSupportRequest.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and organization_id is None:  # noqa: E501
            raise ValueError("Invalid value for `organization_id`, must not be `None`")  # noqa: E501

        self._organization_id = organization_id

    @property
    def requested_at(self):
        """Gets the requested_at of this DecoratedSupportRequest.  # noqa: E501


        :return: The requested_at of this DecoratedSupportRequest.  # noqa: E501
        :rtype: datetime
        """
        return self._requested_at

    @requested_at.setter
    def requested_at(self, requested_at):
        """Sets the requested_at of this DecoratedSupportRequest.


        :param requested_at: The requested_at of this DecoratedSupportRequest.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and requested_at is None:  # noqa: E501
            raise ValueError("Invalid value for `requested_at`, must not be `None`")  # noqa: E501

        self._requested_at = requested_at

    @property
    def expires_at(self):
        """Gets the expires_at of this DecoratedSupportRequest.  # noqa: E501


        :return: The expires_at of this DecoratedSupportRequest.  # noqa: E501
        :rtype: datetime
        """
        return self._expires_at

    @expires_at.setter
    def expires_at(self, expires_at):
        """Sets the expires_at of this DecoratedSupportRequest.


        :param expires_at: The expires_at of this DecoratedSupportRequest.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and expires_at is None:  # noqa: E501
            raise ValueError("Invalid value for `expires_at`, must not be `None`")  # noqa: E501

        self._expires_at = expires_at

    @property
    def requester(self):
        """Gets the requester of this DecoratedSupportRequest.  # noqa: E501


        :return: The requester of this DecoratedSupportRequest.  # noqa: E501
        :rtype: MiniUser
        """
        return self._requester

    @requester.setter
    def requester(self, requester):
        """Sets the requester of this DecoratedSupportRequest.


        :param requester: The requester of this DecoratedSupportRequest.  # noqa: E501
        :type: MiniUser
        """
        if self.local_vars_configuration.client_side_validation and requester is None:  # noqa: E501
            raise ValueError("Invalid value for `requester`, must not be `None`")  # noqa: E501

        self._requester = requester

    @property
    def organization(self):
        """Gets the organization of this DecoratedSupportRequest.  # noqa: E501


        :return: The organization of this DecoratedSupportRequest.  # noqa: E501
        :rtype: MiniOrganization
        """
        return self._organization

    @organization.setter
    def organization(self, organization):
        """Sets the organization of this DecoratedSupportRequest.


        :param organization: The organization of this DecoratedSupportRequest.  # noqa: E501
        :type: MiniOrganization
        """
        if self.local_vars_configuration.client_side_validation and organization is None:  # noqa: E501
            raise ValueError("Invalid value for `organization`, must not be `None`")  # noqa: E501

        self._organization = organization

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DecoratedSupportRequest):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DecoratedSupportRequest):
            return True

        return self.to_dict() != other.to_dict()



import sys

from anyscale.cli_logger import BlockLogger
from anyscale.util import get_wheel_url


def detect_python_minor_version() -> str:
    return f"{sys.version_info[0]}.{sys.version_info[1]}"


def check_required_ray_version(
    logger: BlockLogger,
    ray_version: str,
    ray_commit: str,
    required_ray_version: str,
    required_ray_commit: str,
    ignore_version_check: bool,
) -> None:
    if ray_commit == "{{RAY_COMMIT_SHA}}":
        logger.warning(
            "Ray version checking is skipped because Ray is likely built locally for development. "
            "Compatibility between Ray and Anyscale connect is not guaranteed."
        )
        return

    if ray_version != required_ray_version:
        msg = "The local ray installation has version {}, but {} is required.".format(
            ray_version, required_ray_version
        )
    elif ray_commit != required_ray_commit:
        msg = "The local ray installation has commit {}, but {} is required.".format(
            ray_commit[:7], required_ray_commit[:7]
        )
    else:
        # Version and commit matches.
        return

    msg = (
        "{}\nPlease install the required "
        "Ray version by running:\n\t`pip uninstall ray -y && pip install -U {}`\nTo unsafely "
        "ignore this check, set IGNORE_VERSION_CHECK=1.".format(
            msg, get_wheel_url(required_ray_commit, required_ray_version),
        )
    )
    if ignore_version_check:
        logger.debug(msg)
    else:
        raise ValueError(msg)



from dataclasses import dataclass, field
import re
from tempfile import TemporaryDirectory
from typing import Dict, List, Optional, Union

import click
from click.testing import CliRunner
import pytest

from anyscale._private.docgen.__main__ import generate
from anyscale._private.docgen.generator import MarkdownGenerator, Module
from anyscale._private.models import ModelBase, ModelEnum
from anyscale._private.sdk import sdk_command


def test_full_smoke_test():
    runner = CliRunner()
    with TemporaryDirectory() as d:
        result = runner.invoke(generate, [d])

        assert result.exit_code == 0, result.output
        assert result.output.count("Writing output file") == 4


def _fake_module(
    *,
    name: str = "Fake",
    models: Optional[List] = None,
    cli_commands: Optional[List] = None,
    sdk_commands: Optional[List] = None,
) -> Module:
    return Module(
        title=name,
        filename=f"{name.lower()}.md",
        cli_prefix=f"anyscale {name.lower()}",
        cli_commands=cli_commands or [],
        sdk_prefix=f"anyscale.{name.lower()}",
        sdk_commands=sdk_commands or [],
        models=models or [],
    )


class TestGenerateModelType:
    def test_basic(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            foo: str = field(metadata={"docstring": "Foo docstring."})

            # Private fields should not be documented.
            _private: str

            __doc_yaml_example__ = """\
config: file"""

            __doc_py_example__ = """\
run_python_command()
"""

        output = MarkdownGenerator(
            modules=[_fake_module(models=[FakeConfig])]
        ).generate()
        assert len(output) == 2
        assert "## `FakeConfig`" in output["fake.md"]
        assert "FakeConfig does really special things." in output["fake.md"]
        assert "**`foo` (str)**: Foo docstring." in output["fake.md"]
        assert "_private" not in output["fake.md"]

        # Check that model methods are added.
        assert "def __init__(self, **fields) -> FakeConfig" in output["fake.md"]
        assert "def options(self, **fields) -> FakeConfig" in output["fake.md"]
        assert "def to_dict(self) -> Dict[str, Any]" in output["fake.md"]

        # Check that examples were generated.
        assert '<TabItem value="yamlconfig" label="YAML">' in output["fake.md"]
        assert "config: file" in output["fake.md"]
        assert '<TabItem value="pythonsdk" label="Python">' in output["fake.md"]
        assert "run_python_command()" in output["fake.md"]

    def test_missing_field_docstring(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            foo: str

        with pytest.raises(
            ValueError,
            match="Model 'FakeConfig' is missing a docstring for field 'foo'",
        ):
            MarkdownGenerator(modules=[_fake_module(models=[FakeConfig])]).generate()

    def test_missing_py_example(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            __doc_yaml_example__ = "fake: config"

        with pytest.raises(
            ValueError, match="Model 'FakeConfig' is missing a '__doc_py_example__'",
        ):
            MarkdownGenerator(modules=[_fake_module(models=[FakeConfig])]).generate()

    def test_invalid_py_example(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            __doc_yaml_example__ = "fake: config"
            __doc_py_example__ = "FakeConfig('hi': 123)"

        with pytest.raises(
            ValueError,
            match="'FakeConfig.__doc_py_example__' is not valid Python syntax",
        ):
            MarkdownGenerator(modules=[_fake_module(models=[FakeConfig])]).generate()

    def test_missing_yaml_example(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            __doc_py_example__ = "FakeConfig()"

        with pytest.raises(
            ValueError,
            match="Config model 'FakeConfig' is missing a '__doc_yaml_example__'",
        ):
            MarkdownGenerator(modules=[_fake_module(models=[FakeConfig])]).generate()

    def test_invalid_yaml_example(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            __doc_yaml_example__ = "fake: config: oops"
            __doc_py_example__ = "FakeConfig()"

        with pytest.raises(
            ValueError,
            match="'FakeConfig.__doc_yaml_example__' is not valid YAML syntax",
        ):
            MarkdownGenerator(modules=[_fake_module(models=[FakeConfig])]).generate()

    def test_non_config_does_not_have_constructors(self):
        """Models that don't end in "Config" shouldn't document constructor methods."""

        @dataclass(frozen=True)
        class FakeStatus(ModelBase):
            """FakeStatus is so cool."""

            foo: str = field(metadata={"docstring": "Foo docstring."})

            __doc_py_example__ = "FakeStatus()"
            __doc_yaml_example__ = "foo: hi"

        output = MarkdownGenerator(
            modules=[_fake_module(models=[FakeStatus])]
        ).generate()

        assert "def __init__(self, **fields)" not in output["fake.md"]
        assert "def options(self, **fields)" not in output["fake.md"]
        assert "def to_dict(self) -> Dict[str, Any]" in output["fake.md"]

    def test_type_hints(self):
        @dataclass(frozen=True)
        class FakeEnum(ModelEnum):
            pass

        @dataclass(frozen=True)
        class FakeStatus(ModelBase):
            """FakeStatus is so cool."""

            __doc_py_example__ = "FakeConfig()"

        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            foo: str = field(metadata={"docstring": "foo docs"})
            bar: Optional[str] = field(metadata={"docstring": "bar docs"})
            baz: Dict[str, FakeStatus] = field(metadata={"docstring": "baz docs"})
            qux: List[Optional[FakeEnum]] = field(metadata={"docstring": "qux docs"})

            __doc_py_example__ = "FakeConfig()"
            __doc_yaml_example__ = "fake: config"

        output = MarkdownGenerator(
            modules=[_fake_module(models=[FakeConfig, FakeStatus, FakeEnum])]
        ).generate()

        assert "**`foo` (str)**: foo docs" in output["fake.md"]
        assert "**`bar` (str | None)**: bar docs" in output["fake.md"]
        assert (
            "**`baz` (Dict[str, [FakeStatus](fake#fakestatus)])**: baz docs"
            in output["fake.md"]
        )
        assert (
            "**`qux` (List[[FakeEnum](fake#fakeenum) | None])**: qux docs"
            in output["fake.md"]
        )

    def test_customer_hosted_cloud_only(self):
        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does special things."""

            # Should not be labeled as customer-hosted cloud only.
            foo: str = field(metadata={"docstring": "Foo docstring."})

            __doc_yaml_example__ = """\
config: file"""

            __doc_py_example__ = """\
run_python_command()
"""

        output = MarkdownGenerator(
            modules=[_fake_module(models=[FakeConfig])]
        ).generate()

        # By default options should not be labeled as customer-hosted cloud only.
        assert "Only available on [customer-hosted clouds]" not in output["fake.md"]

        @dataclass(frozen=True)
        class FakeConfigTwo(ModelBase):
            """FakeConfigTwo does really special things."""

            # Should be labeled as customer-hosted cloud only.
            foo: str = field(
                metadata={"docstring": "Foo docstring.", "customer_hosted_only": True}
            )

            __doc_yaml_example__ = """\
config: file"""

            __doc_py_example__ = """\
run_python_command()
"""

        output = MarkdownGenerator(
            modules=[_fake_module(models=[FakeConfigTwo])]
        ).generate()

        assert "Only available on [customer-hosted clouds]" in output["fake.md"]


class TestGenerateModelEnumType:
    def test_basic(self):
        @dataclass(frozen=True)
        class FakeEnum(ModelEnum):
            """FakeEnum is highly important."""

            FOO = "FOO"
            BAR = "BAR"

            __docstrings__ = {
                "FOO": "Foo is quite important.",
                "BAR": "But bar is even more so.",
            }

        output = MarkdownGenerator(modules=[_fake_module(models=[FakeEnum])]).generate()
        assert len(output) == 2
        assert "FakeEnum is highly important." in output["fake.md"]
        assert "**`FOO`**: Foo is quite important." in output["fake.md"]
        assert "**`BAR`**: But bar is even more so." in output["fake.md"]

    def test_reject_value_missing_docstring(self):
        with pytest.raises(
            ValueError,
            match=re.escape(
                "ModelEnum 'FakeEnum.__docstrings__' is missing docstrings for values: ['BAR', 'FOO']"
            ),
        ):

            @dataclass(frozen=True)
            class FakeEnum(ModelEnum):
                """FakeEnum is highly important."""

                FOO = "FOO"
                BAR = "BAR"

        with pytest.raises(
            ValueError,
            match=re.escape(
                "ModelEnum 'FakeEnumTwo.__docstrings__' is missing docstrings for values: ['BAR']"
            ),
        ):

            @dataclass(frozen=True)
            class FakeEnumTwo(ModelEnum):
                """FakeEnumTwo is also important."""

                FOO = "FOO"
                BAR = "BAR"

                __docstrings__ = {
                    "FOO": "Foo is quite important.",
                }


class TestGenerateCLICommand:
    def test_basic(self):
        @click.command("do-something")
        @click.argument("the-arg")
        @click.option("-m", "--maybe", help="Maybe do something else.")
        def do_something():
            """This is the description of what to do."""

        output = MarkdownGenerator(
            modules=[_fake_module(cli_commands=[do_something])]
        ).generate()
        assert "## `anyscale fake do-something`" in output["fake.md"]
        assert "anyscale fake do-something [OPTIONS] THE_ARG" in output["fake.md"]
        assert "This is the description of what to do." in output["fake.md"]
        assert "**`-m/--maybe`**: Maybe do something else." in output["fake.md"]
        assert "--help" not in output["fake.md"]


class TestGenerateSDKCommand:
    def test_basic(self):
        @sdk_command(
            "TEST_KEY",
            None,
            doc_py_example="do_something()",
            arg_docstrings={"s": "a string"},
        )
        def do_something(s: str, *, _sdk) -> int:
            """Might do something."""

        output = MarkdownGenerator(
            modules=[_fake_module(sdk_commands=[do_something])]
        ).generate()

        assert "## `anyscale.fake.do_something`" in output["fake.md"]
        assert "Might do something." in output["fake.md"]
        assert "**`s` (str)**: a string" in output["fake.md"]
        assert "**Returns**: int" in output["fake.md"]
        # Private arguments should not be documented.
        assert "_sdk" not in output["fake.md"]

    def test_no_return(self):
        @sdk_command(
            "TEST_KEY",
            None,
            doc_py_example="do_something()",
            arg_docstrings={"s": "a string"},
        )
        def do_something(s: str):
            """Might do something."""

        output = MarkdownGenerator(
            modules=[_fake_module(sdk_commands=[do_something])]
        ).generate()

        assert "Returns" not in output["fake.md"]

    def test_missing_docstring(self):
        @sdk_command(
            "TEST_KEY",
            None,
            doc_py_example="do_something()",
            arg_docstrings={"s": "a string"},
        )
        def do_something(s: str, *, _sdk):
            pass

        with pytest.raises(
            ValueError,
            match="SDK command 'anyscale.fake.do_something' is missing a docstring.",
        ):
            MarkdownGenerator(
                modules=[_fake_module(sdk_commands=[do_something])]
            ).generate()

    def test_missing_py_example(self):
        with pytest.raises(
            ValueError,
            match="SDK command 'do_something' must provide a non-empty 'doc_py_example'.",
        ):

            @sdk_command(
                "TEST_KEY", None, doc_py_example="", arg_docstrings={"s": "a string"},
            )
            def do_something(s: str, *, _sdk):
                """Might do something."""

    def test_missing_arg_docstring(self):
        @sdk_command(
            "TEST_KEY", None, doc_py_example="do_something()", arg_docstrings={},
        )
        def do_something(s: str, *, _sdk):
            """Might do something."""

        with pytest.raises(
            ValueError,
            match="SDK command 'anyscale.fake.do_something' is missing a docstring for argument 's'",
        ):
            MarkdownGenerator(
                modules=[_fake_module(sdk_commands=[do_something])]
            ).generate()

    def test_type_hints(self):
        """Models that don't end in "Config" shouldn't document constructor methods."""

        @dataclass(frozen=True)
        class FakeEnum(ModelEnum):
            pass

        @dataclass(frozen=True)
        class FakeStatus(ModelBase):
            """FakeStatus is so cool."""

            __doc_py_example__ = "FakeStatus()"

        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            __doc_py_example__ = "FakeConfig()"
            __doc_yaml_example__ = "fake: config"

        @sdk_command(
            "TEST_KEY",
            None,
            doc_py_example="do_something()",
            arg_docstrings={
                "foo": "foo docs",
                "bar": "bar docs",
                "baz": "baz docs",
                "qux": "qux docs",
                "quux": "quux docs",
            },
        )
        def do_something(
            foo: FakeEnum,
            bar: Optional[FakeStatus],
            baz: Union[Optional[FakeConfig], FakeStatus],
            qux: List[FakeEnum],
            quux: Dict[str, Optional[FakeEnum]],
        ) -> Optional[FakeStatus]:
            """Might do something."""

        output = MarkdownGenerator(
            modules=[
                _fake_module(
                    sdk_commands=[do_something],
                    models=[FakeEnum, FakeStatus, FakeConfig],
                )
            ]
        ).generate()

        assert "**`foo` ([FakeEnum](fake#fakeenum))**: foo docs" in output["fake.md"]
        assert (
            "**`bar` ([FakeStatus](fake#fakestatus) | None)**: bar docs"
            in output["fake.md"]
        )
        assert (
            "**`baz` ([FakeConfig](fake#fakeconfig) | None | [FakeStatus](fake#fakestatus))**: baz docs"
            in output["fake.md"]
        )
        assert (
            "**`qux` (List[[FakeEnum](fake#fakeenum)])**: qux docs" in output["fake.md"]
        )
        assert (
            "**`quux` (Dict[str, [FakeEnum](fake#fakeenum) | None])**: quux docs"
            in output["fake.md"]
        )
        assert "**Returns**: [FakeStatus](fake#fakestatus) | None" in output["fake.md"]


class TestIndex:
    def test_basic(self):
        @click.command("do-something")
        @click.argument("the-arg")
        @click.option("-m", "--maybe", help="Maybe do something else.")
        def do_something_cli():
            """This is the description of what to do."""

        @dataclass(frozen=True)
        class FakeConfig(ModelBase):
            """FakeConfig does really special things."""

            __doc_py_example__ = "FakeConfig()"
            __doc_yaml_example__ = "fake: config"

        @sdk_command(
            "TEST_KEY",
            None,
            doc_py_example="do_something()",
            arg_docstrings={"foo": "foo docs"},
        )
        def do_something_sdk(foo: str):
            """Might do something."""

        output = MarkdownGenerator(
            modules=[
                _fake_module(
                    cli_commands=[do_something_cli],
                    sdk_commands=[do_something_sdk],
                    models=[FakeConfig],
                ),
                _fake_module(
                    name="Other",
                    cli_commands=[do_something_cli],
                    sdk_commands=[do_something_sdk],
                    models=[FakeConfig],
                ),
            ]
        ).generate()
        assert set(output.keys()) == {"index.md", "fake.md", "other.md"}

        assert (
            output["index.md"]
            == """\
# Platform API

- [Fake API Reference](fake)
  - [Fake Models](fake#fake-models)
  - [Fake CLI](fake#fake-cli)
  - [Fake SDK](fake#fake-sdk)
- [Other API Reference](other)
  - [Other Models](other#other-models)
  - [Other CLI](other#other-cli)
  - [Other SDK](other#other-sdk)
"""
        )



"""Entrypoint to generate reference docs for the Anyscale CLI/SDK.

Usage: python -m anyscale._private.docgen --help
"""
import os

import click

import anyscale
from anyscale._private.docgen.generator import MarkdownGenerator, Module
from anyscale.commands import compute_config_commands, job_commands, service_commands
from anyscale.compute_config.models import (
    ComputeConfig,
    ComputeConfigVersion,
    HeadNodeConfig,
    MarketType,
    WorkerNodeGroupConfig,
)
from anyscale.job.models import (
    JobConfig,
    JobRunState,
    JobRunStatus,
    JobState,
    JobStatus,
)
from anyscale.service.models import (
    RayGCSExternalStorageConfig,
    ServiceConfig,
    ServiceState,
    ServiceStatus,
    ServiceVersionState,
    ServiceVersionStatus,
)


# Defines all modules to be documented.
ALL_MODULES = [
    Module(
        title="Job",
        filename="job-api.md",
        cli_prefix="anyscale job",
        cli_commands=[
            job_commands.submit,
            job_commands.status,
            job_commands.terminate,
            job_commands.archive,
        ],
        sdk_prefix="anyscale.job",
        sdk_commands=[
            anyscale.job.submit,
            anyscale.job.status,
            anyscale.job.terminate,
            anyscale.job.archive,
        ],
        models=[JobConfig, JobState, JobStatus, JobRunStatus, JobRunState],
    ),
    Module(
        title="Service",
        filename="service-api.md",
        cli_prefix="anyscale service",
        cli_commands=[
            service_commands.deploy,
            service_commands.status,
            service_commands.wait,
            service_commands.rollback,
            service_commands.terminate,
        ],
        sdk_prefix="anyscale.service",
        sdk_commands=[
            anyscale.service.deploy,
            anyscale.service.status,
            anyscale.service.wait,
            anyscale.service.rollback,
            anyscale.service.terminate,
        ],
        models=[
            ServiceConfig,
            RayGCSExternalStorageConfig,
            ServiceStatus,
            ServiceState,
            ServiceVersionStatus,
            ServiceVersionState,
        ],
    ),
    Module(
        title="Compute Config",
        filename="compute-config-api.md",
        cli_prefix="anyscale compute-config",
        cli_commands=[
            compute_config_commands.create_compute_config,
            compute_config_commands.get_compute_config,
            compute_config_commands.archive_compute_config,
        ],
        sdk_prefix="anyscale.compute_config",
        sdk_commands=[
            anyscale.compute_config.create,
            anyscale.compute_config.get,
            anyscale.compute_config.archive,
        ],
        models=[
            ComputeConfig,
            HeadNodeConfig,
            WorkerNodeGroupConfig,
            MarketType,
            ComputeConfigVersion,
        ],
    ),
]


@click.command(help="Generate markdown docs for the Anyscale CLI & SDK.")
@click.argument("output_dir")
@click.option(
    "-r",
    "--remove-existing",
    is_flag=True,
    default=False,
    help="If set, all files in the 'output_dir' that were not generated will be removed.",
)
def generate(
    output_dir: str, *, remove_existing: bool = False,
):
    if not os.path.isdir(output_dir):
        raise RuntimeError(f"output_dir '{output_dir}' does not exist.")

    gen = MarkdownGenerator(ALL_MODULES)
    gen.generate()

    generated_files = set()
    os.makedirs(output_dir, exist_ok=True)
    for filename, content in gen.generate().items():
        generated_files.add(filename)
        full_path = os.path.join(output_dir, filename)
        print(f"Writing output file {full_path}")
        with open(full_path, "w") as f:
            f.write(content)

    if remove_existing:
        to_remove = set(os.listdir(output_dir)) - generated_files
        for path in to_remove:
            full_path = os.path.join(output_dir, path)
            print(f"Removing existing file {full_path}")
            os.unlink(full_path)


if __name__ == "__main__":
    generate()



from datetime import datetime
from typing import Generic, List, Optional, TypeVar
from unittest.mock import Mock, patch

import click
import pytest

from anyscale.anyscale_pydantic import BaseModel
from anyscale.client.openapi_client.models import (
    AnyscaleServiceAccount,
    AnyscaleserviceaccountResponse,
    OrganizationCollaborator,
    OrganizationPermissionLevel,
    ServerSessionToken,
    ServersessiontokenResponse,
)
from anyscale.controllers.service_account_controller import ServiceAccountController


T = TypeVar("T")


class ListResponseMetadata(BaseModel):
    total: Optional[int] = None
    next_paging_token: Optional[str] = None


class ListResponse(BaseModel, Generic[T]):
    results: List[T]
    metadata: ListResponseMetadata


test_service_account_0 = AnyscaleServiceAccount(
    email="test_email",
    user_id="test_user_id",
    name="test_name",
    organization_id="test_org_id",
    permission_level=OrganizationPermissionLevel.COLLABORATOR,
    created_at=datetime(2021, 1, 1),
)

test_service_account_1 = AnyscaleServiceAccount(
    email="test_email_1",
    user_id="test_user_id_1",
    name="test_name_1",
    organization_id="test_org_id_1",
    permission_level=OrganizationPermissionLevel.COLLABORATOR,
    created_at=datetime(2021, 1, 1),
)

test_organization_collaborator_0 = OrganizationCollaborator(
    id="test_identity_id",
    name=test_service_account_0.name,
    permission_level=test_service_account_0.permission_level,
    created_at=test_service_account_0.created_at,
    email=test_service_account_0.email,
    user_id=test_service_account_0.user_id,
)

test_organization_collaborator_1 = OrganizationCollaborator(
    id="test_identity_id_1",
    name=test_service_account_1.name,
    permission_level=test_service_account_1.permission_level,
    created_at=test_service_account_1.created_at,
    email=test_service_account_1.email,
    user_id=test_service_account_1.user_id,
)


@pytest.fixture()
def mock_api_client() -> Mock:
    mock_api_client = Mock()
    mock_api_client.create_service_account_api_v2_users_service_accounts_post = Mock(
        return_value=AnyscaleserviceaccountResponse(result=test_service_account_0)
    )
    mock_api_client.create_api_key_api_v2_users_create_api_key_post = Mock(
        return_value=ServersessiontokenResponse(
            result=ServerSessionToken(server_session_id="test_server_session_id")
        )
    )

    def list_organization_collaborators_api_v2_organization_collaborators_get(
        is_service_account=None, email=None
    ):
        if email == "test_email":
            return ListResponse(
                results=[test_organization_collaborator_0],
                metadata=ListResponseMetadata(total=1),
            )
        if email == "test_email_1":
            return ListResponse(
                results=[test_organization_collaborator_1],
                metadata=ListResponseMetadata(total=1),
            )
        if email == "non_existent_email":
            return ListResponse(results=[], metadata=ListResponseMetadata(total=0))
        return ListResponse(
            results=[
                test_organization_collaborator_0,
                test_organization_collaborator_1,
            ],
            metadata=ListResponseMetadata(total=2),
        )

    mock_api_client.list_organization_collaborators_api_v2_organization_collaborators_get = Mock(
        side_effect=list_organization_collaborators_api_v2_organization_collaborators_get
    )
    return mock_api_client


@pytest.fixture()
def mock_auth_api_client(mock_api_client: Mock, base_mock_anyscale_api_client: Mock):
    mock_auth_api_client = Mock(
        api_client=mock_api_client, anyscale_api_client=base_mock_anyscale_api_client,
    )
    with patch.multiple(
        "anyscale.controllers.base_controller",
        get_auth_api_client=Mock(return_value=mock_auth_api_client),
    ):
        yield


def test_create_service_account(mock_api_client, mock_auth_api_client):
    sa_controller = ServiceAccountController()
    mock_console = Mock()
    sa_controller.console = mock_console
    sa_controller.api_client = mock_api_client
    sa_controller.create_service_account("test_name")

    mock_api_client.create_service_account_api_v2_users_service_accounts_post.assert_called_once_with(
        name="test_name"
    )
    mock_api_client.create_api_key_api_v2_users_create_api_key_post.assert_called_once_with(
        api_key_parameters={"user_id": "test_user_id", "duration": 3153600000}
    )

    mock_console.print.assert_called_with("test_server_session_id")


def test_create_new_service_account_api_key(mock_api_client, mock_auth_api_client):
    sa_controller = ServiceAccountController()
    mock_console = Mock()
    sa_controller.console = mock_console
    sa_controller.api_client = mock_api_client
    sa_controller.create_new_service_account_api_key("test_email")

    mock_api_client.list_organization_collaborators_api_v2_organization_collaborators_get.assert_called_once_with(
        is_service_account=True, email="test_email"
    )
    mock_api_client.create_api_key_api_v2_users_create_api_key_post.assert_called_once_with(
        api_key_parameters={"user_id": "test_user_id", "duration": 3153600000}
    )
    mock_console.print.assert_called_with("test_server_session_id")


@pytest.mark.parametrize("max_items", [1, 2])
def test_list_service_accounts(mock_api_client, mock_auth_api_client, max_items):
    sa_controller = ServiceAccountController()
    sa_controller._print_service_account_table = Mock()
    mock_console = Mock()
    sa_controller.console = mock_console
    sa_controller.api_client = mock_api_client
    sa_controller.list_service_accounts(max_items)

    mock_api_client.list_organization_collaborators_api_v2_organization_collaborators_get.assert_called_once_with(
        is_service_account=True
    )
    if max_items == 1:
        sa_controller._print_service_account_table.assert_called_once_with(
            [test_organization_collaborator_0]
        )
    else:
        sa_controller._print_service_account_table.assert_called_once_with(
            [test_organization_collaborator_0, test_organization_collaborator_1]
        )


@pytest.mark.parametrize(
    "input_email", ["test_email", "non_existent_email", "multiple_service_accounts"]
)
def test_delete_service_account(mock_api_client, mock_auth_api_client, input_email):
    sa_controller = ServiceAccountController()
    mock_console = Mock()
    sa_controller.console = mock_console
    sa_controller.api_client = mock_api_client

    if input_email == "non_existent_email":
        with pytest.raises(
            click.ClickException,
            match=f"No service account with email {input_email} found.",
        ):
            sa_controller.delete_service_account(input_email)
        mock_api_client.list_organization_collaborators_api_v2_organization_collaborators_get.assert_called_once_with(
            is_service_account=True, email=input_email
        )
        return

    if input_email == "multiple_service_accounts":
        with pytest.raises(
            click.ClickException,
            match=f"Internal server error when fetching service account with email {input_email}. Please contact support.",
        ):
            sa_controller.delete_service_account(input_email)
        mock_api_client.list_organization_collaborators_api_v2_organization_collaborators_get.assert_called_once_with(
            is_service_account=True, email=input_email
        )
        return

    sa_controller.delete_service_account(input_email)
    mock_api_client.list_organization_collaborators_api_v2_organization_collaborators_get.assert_called_once_with(
        is_service_account=True, email=input_email
    )
    mock_api_client.remove_organization_collaborator_api_v2_organization_collaborators_identity_id_delete(
        user_id="test_user_id"
    )


def test_rotate_service_account_api_keys(mock_api_client, mock_auth_api_client):
    sa_controller = ServiceAccountController()
    mock_console = Mock()
    sa_controller.console = mock_console
    sa_controller.api_client = mock_api_client

    sa_controller.rotate_service_account_api_keys("test_email")
    mock_api_client.list_organization_collaborators_api_v2_organization_collaborators_get.assert_called_once_with(
        is_service_account=True, email="test_email"
    )
    mock_api_client.create_api_key_api_v2_users_create_api_key_post.assert_called_once_with(
        api_key_parameters={"user_id": "test_user_id", "duration": 3153600000}
    )
    mock_api_client.rotate_api_key_for_user_api_v2_organization_collaborators_rotate_api_key_for_user_user_id_post.assert_called_once_with(
        "test_user_id"
    )
    mock_console.print.assert_called_with("test_server_session_id")



import os
from subprocess import list2cmdline
from typing import Generator, List, Optional, Union
from unittest.mock import Mock, patch
import uuid

import click
from click.testing import CliRunner
import pytest

from anyscale._private.sdk import _LAZY_SDK_SINGLETONS
from anyscale.commands.job_commands import status, submit, wait
from anyscale.job.commands import _JOB_SDK_SINGLETON_KEY
from anyscale.job.models import (
    JobConfig,
    JobRunState,
    JobRunStatus,
    JobState,
    JobStatus,
)


def _get_test_file_path(subpath: str) -> str:
    return os.path.join(os.path.dirname(__file__), "test_files/", subpath)


EMPTY_CONFIG_PATH = _get_test_file_path("job_config_files/empty.yaml")
MINIMAL_CONFIG_PATH = _get_test_file_path("job_config_files/minimal.yaml")
NON_YAML_CONFIG_PATH = _get_test_file_path("job_config_files/minimal_job_config")
NONEXISTENT_CONFIG_PATH = _get_test_file_path("job_config_files/nonexistent.yaml")
FULL_CONFIG_PATH = _get_test_file_path("job_config_files/full.yaml")
POINTS_TO_REQUIREMENTS_FILE_PATH = _get_test_file_path(
    "job_config_files/points_to_requirements_file.yaml"
)
UNRECOGNIZED_OPTION_CONFIG_PATH = _get_test_file_path(
    "job_config_files/unrecognized_option.yaml"
)

COMMENTS_REQ_PATH = _get_test_file_path("requirements_files/comments.txt")
EMPTY_REQ_PATH = _get_test_file_path("requirements_files/empty.txt")
MULIT_LINE_WITH_WHITESPACE__REQ_PATH = _get_test_file_path(
    "requirements_files/multi_line_with_whitespace.txt"
)
MULTI_LINE_REQ_PATH = _get_test_file_path("requirements_files/multi_line.txt")
SINGLE_LINE_REQ_PATH = _get_test_file_path("requirements_files/single_line.txt")
WORKSPACE_REQUIREMENTS_REQ_PATH = _get_test_file_path(
    "requirements_files/test_workspace_requirements.txt"
)

SINGLE_EXCLUDE_LIST = ["torch"]
MULTIPLE_EXCLUDE_LIST = ["numpy", "pandas", "scipy"]

DEFAULT_JOB_ID = "default-fake-job-id"


class FakeJobSDK:
    DEFAULT_JOB_NAME = "default-fake-job-name"
    DEFAULT_JOB_CONFIG = JobConfig(entrypoint="python hello.py")

    def __init__(self):
        self.submitted_config: Optional[JobConfig] = None
        self.submitted_id: Optional[str] = None
        self.submitted_name: Optional[str] = None
        self.fetched_name: Optional[str] = None
        self.waited_id: Optional[str] = None
        self.waited_name: Optional[str] = None
        self.waited_state: Optional[Union[str, JobState]] = None
        self.waited_timeout_s: Optional[float] = None
        self._job_runs: List[JobRunStatus] = []

    def submit(self, config: JobConfig):
        assert isinstance(config, JobConfig)
        self.submitted_config = config
        self.submitted_id = str(uuid.uuid4())
        self.submitted_name = (
            self.submitted_config.name
            if self.submitted_config.name is not None
            else self.DEFAULT_JOB_NAME
        )
        return self.submitted_id

    def wait(
        self,
        name: Optional[str] = None,
        job_id: Optional[str] = None,
        state: Union[str, JobState] = JobState.SUCCEEDED,
        timeout_s: int = 1800,
    ) -> str:
        self.waited_id = job_id or str(uuid.uuid4())
        self.waited_name = name
        self.waited_state = state
        self.waited_timeout_s = timeout_s
        return self.waited_id

    def status(self, name: Optional[str], job_id: Optional[str]):
        self.fetched_name = name
        if (
            self.submitted_config is not None
            and self.submitted_name is not None
            and (name == self.submitted_name or job_id == self.submitted_id)
        ):
            return JobStatus(
                id=self.submitted_id if self.submitted_id else "",
                name=self.submitted_name,
                state=JobState.SUCCEEDED,
                config=self.submitted_config,
                runs=self._job_runs,
            )
        raise RuntimeError("Job was not found.")


@pytest.fixture()
def mock_job_controller() -> Generator[Mock, None, None]:
    mock_job_controller = Mock(submit=Mock(return_value=DEFAULT_JOB_ID))
    mock_job_controller_cls = Mock(return_value=mock_job_controller)
    with patch(
        "anyscale.commands.job_commands.JobController", new=mock_job_controller_cls,
    ):
        yield mock_job_controller


@pytest.fixture()
def fake_job_sdk() -> Generator[FakeJobSDK, None, None]:
    fake_job_sdk = FakeJobSDK()
    _LAZY_SDK_SINGLETONS[_JOB_SDK_SINGLETON_KEY] = fake_job_sdk
    try:
        yield fake_job_sdk
    finally:
        del _LAZY_SDK_SINGLETONS[_JOB_SDK_SINGLETON_KEY]


def _assert_error_message(result: click.testing.Result, *, message: str):
    assert result.exit_code != 0
    assert message in result.stdout


class TestSubmit:
    def test_missing_arg(self, fake_job_sdk, mock_job_controller):
        runner = CliRunner()
        result = runner.invoke(submit)
        _assert_error_message(
            result,
            message="Either a config file or an inlined entrypoint must be provided.",
        )

    def test_config_file_not_found(self, fake_job_sdk, mock_job_controller):
        runner = CliRunner()
        result = runner.invoke(submit, ["missing_config.yaml"])
        _assert_error_message(
            result, message="Job config file 'missing_config.yaml' not found.",
        )

    @pytest.mark.parametrize("inlined_entrypoint", [False, True])
    def test_basic(self, fake_job_sdk, mock_job_controller, inlined_entrypoint: bool):
        runner = CliRunner()
        entrypoint_args = (
            ["--", "python", "main.py"] if inlined_entrypoint else [MINIMAL_CONFIG_PATH]
        )
        result = runner.invoke(submit, [*entrypoint_args])
        assert result.exit_code == 0, result.stdout

        if inlined_entrypoint:
            # Passing an inlined entrypoint should go through the new path.
            assert fake_job_sdk.submitted_config == JobConfig(
                entrypoint="python main.py"
            )
            mock_job_controller.submit.assert_not_called()
        else:
            # For now, passing a config file should go through the old path.
            assert fake_job_sdk.submitted_config is None
            mock_job_controller.submit.assert_called_once_with(
                MINIMAL_CONFIG_PATH, name=None
            )

    @pytest.mark.parametrize(
        "config_file_arg",
        [MINIMAL_CONFIG_PATH, NON_YAML_CONFIG_PATH, NONEXISTENT_CONFIG_PATH,],
    )
    def test_inlined_config_file(
        self, fake_job_sdk, mock_job_controller, config_file_arg
    ):
        runner = CliRunner()
        result = runner.invoke(submit, [config_file_arg])

        if config_file_arg == NONEXISTENT_CONFIG_PATH:
            assert result.exit_code == 1
            error_msg = f"Job config file '{config_file_arg}' not found"
            assert error_msg in result.stdout
        else:
            assert result.exit_code == 0, result.stdout

            mock_job_controller.submit.assert_called_once_with(
                config_file_arg, name=None
            )
        assert fake_job_sdk.submitted_config is None

    @pytest.mark.parametrize(
        "config_file",
        [
            EMPTY_CONFIG_PATH,
            MINIMAL_CONFIG_PATH,
            FULL_CONFIG_PATH,
            POINTS_TO_REQUIREMENTS_FILE_PATH,
            UNRECOGNIZED_OPTION_CONFIG_PATH,
        ],
    )
    @pytest.mark.parametrize(
        "override",
        [
            [],
            ["--name", "test-name"],
            ["--image-uri", "user_name/repository"],
            ["--containerfile", "somefile"],
            ["--working-dir", "."],
            ["--", "python", "main.py"],
        ],
    )
    def test_config_file(
        self, fake_job_sdk, mock_job_controller, config_file: str, override: List[str],
    ):
        runner = CliRunner()
        command = ["--config-file", config_file, *override]
        result = runner.invoke(submit, command)

        if config_file == UNRECOGNIZED_OPTION_CONFIG_PATH:
            assert result.exit_code != 0
            return

        if config_file == EMPTY_CONFIG_PATH and "--" not in override:
            # When the config file doesn't have the entrypoint,
            # entrypoint must be provided as a CLI argument.
            assert result.exit_code != 0
            return

        assert result.exit_code == 0, result.stdout

        if config_file == EMPTY_CONFIG_PATH:
            assert override[0] == "--"
            expected = JobConfig(entrypoint=list2cmdline(override[1:]))
        else:
            expected = JobConfig.from_yaml(config_file)

        if override:
            if override[0] == "--name":
                expected = expected.options(name=override[1])
            elif override[0] == "--image-uri":
                expected = expected.options(image_uri=override[1])
            elif override[0] == "--containerfile":
                expected = expected.options(containerfile=override[1])
            elif override[0] == "--working-dir":
                expected = expected.options(working_dir=override[1])
            elif override[0] == "--":
                expected = expected.options(entrypoint=list2cmdline(override[1:]))

        assert fake_job_sdk.submitted_config == expected
        mock_job_controller.submit.assert_not_called()

    def test_submit_both_image_uri_and_containerfile(
        self, fake_job_sdk, mock_job_controller
    ):
        runner = CliRunner()
        result = runner.invoke(
            submit,
            [
                "--config-file",
                FULL_CONFIG_PATH,
                "--image-uri",
                "image",
                "--containerfile",
                "file",
            ],
        )
        _assert_error_message(
            result,
            message="Only one of '--containerfile' and '--image-uri' can be provided.",
        )

    @pytest.mark.parametrize("inlined_entrypoint", [False, True])
    def test_override_name(
        self, fake_job_sdk, mock_job_controller, inlined_entrypoint: bool
    ):
        runner = CliRunner()
        entrypoint_args = (
            ["--", "python", "main.py"] if inlined_entrypoint else [MINIMAL_CONFIG_PATH]
        )
        result = runner.invoke(submit, ["--name", "test-name", *entrypoint_args])
        assert result.exit_code == 0

        if inlined_entrypoint:
            assert fake_job_sdk.submitted_config == JobConfig(
                entrypoint="python main.py", name="test-name"
            )
            mock_job_controller.submit.assert_not_called()
        else:
            assert fake_job_sdk.submitted_config is None
            mock_job_controller.submit.assert_called_once_with(
                MINIMAL_CONFIG_PATH, name="test-name"
            )

    @pytest.mark.parametrize("inlined_entrypoint", [False, True])
    def test_no_wait_by_default(
        self, fake_job_sdk, mock_job_controller, inlined_entrypoint: bool
    ):
        runner = CliRunner()

        entrypoint_args = (
            ["--", "python", "main.py"] if inlined_entrypoint else [MINIMAL_CONFIG_PATH]
        )
        result = runner.invoke(submit, [*entrypoint_args])
        assert result.exit_code == 0

        assert fake_job_sdk.waited_id is None

    @pytest.mark.parametrize("inlined_entrypoint", [False, True])
    def test_wait(self, fake_job_sdk, mock_job_controller, inlined_entrypoint: bool):
        runner = CliRunner()

        entrypoint_args = (
            ["--", "python", "main.py"] if inlined_entrypoint else [MINIMAL_CONFIG_PATH]
        )
        result = runner.invoke(submit, ["--wait", *entrypoint_args])
        assert result.exit_code == 0
        print(result.stdout)

        if inlined_entrypoint:
            assert (
                fake_job_sdk.waited_id is not None
                and fake_job_sdk.waited_id == fake_job_sdk.submitted_id
            )
        else:
            assert (
                fake_job_sdk.waited_id is not None
                and fake_job_sdk.waited_id == DEFAULT_JOB_ID
            )

    def test_missing_requirements(self, fake_job_sdk, mock_job_controller):
        runner = CliRunner()
        entrypoint_args = ["--", "python", "main.py"]
        missing_req_file = "not-real-req.txt"
        result = runner.invoke(submit, ["-r", missing_req_file, *entrypoint_args])

        assert result.exit_code == 1
        error_msg = f"Requirements file '{missing_req_file}' not found"
        assert error_msg in result.stdout

    @pytest.mark.parametrize(
        "requirement_path",
        [
            COMMENTS_REQ_PATH,
            EMPTY_REQ_PATH,
            MULIT_LINE_WITH_WHITESPACE__REQ_PATH,
            MULTI_LINE_REQ_PATH,
            SINGLE_LINE_REQ_PATH,
            WORKSPACE_REQUIREMENTS_REQ_PATH,
        ],
    )
    def test_requirements(self, fake_job_sdk, mock_job_controller, requirement_path):
        runner = CliRunner()
        entrypoint_args = ["--", "python", "main.py"]
        result = runner.invoke(submit, ["-r", requirement_path, *entrypoint_args])
        assert result.exit_code == 0
        assert fake_job_sdk.submitted_config.requirements == requirement_path

    @pytest.mark.parametrize(
        "exclude_list", [SINGLE_EXCLUDE_LIST, MULTIPLE_EXCLUDE_LIST]
    )
    def test_exclude(self, exclude_list, fake_job_sdk, mock_job_controller):
        runner = CliRunner()
        entrypoint_args = ["--", "python", "main.py"]
        exclude_args = [i for exclude in exclude_list for i in ("-e", exclude)]
        result = runner.invoke(submit, [*exclude_args, *entrypoint_args])
        assert result.exit_code == 0
        assert fake_job_sdk.submitted_config.excludes == exclude_list

    def test_compute_config(self, fake_job_sdk, mock_job_controller):
        runner = CliRunner()
        entrypoint_args = ["--", "python", "main.py"]
        compute_config_name = "override-default-compute-config"
        result = runner.invoke(
            submit, ["--compute-config", compute_config_name, *entrypoint_args]
        )
        assert result.exit_code == 0
        assert fake_job_sdk.submitted_config.compute_config == compute_config_name


class TestStatus:
    def test_no_name(self, fake_job_sdk):
        runner = CliRunner()
        result = runner.invoke(status)
        _assert_error_message(
            result, message="One of '--name' and '--job-id' must be provided."
        )

    def test_name_and_job_id(self, fake_job_sdk):
        runner = CliRunner()
        result = runner.invoke(status, ["-n", "custom-name", "--id", "fake-job-id"])
        _assert_error_message(
            result, message="Only one of '--name' and '--job-id' can be provided."
        )

    def test_name(self, fake_job_sdk, mock_job_controller):
        runner = CliRunner()
        runner.invoke(submit, ["-n", "custom-name", "--", "python", "main.py"])
        result = runner.invoke(status, ["-n", "custom-name"])
        assert result.exit_code == 0
        assert fake_job_sdk.submitted_id in result.stdout
        assert fake_job_sdk.submitted_name in result.stdout

    def test_id(self, fake_job_sdk, mock_job_controller):
        runner = CliRunner()
        runner.invoke(submit, ["--", "python", "main.py"])
        result = runner.invoke(status, ["--job-id", fake_job_sdk.submitted_id])
        assert result.exit_code == 0
        assert fake_job_sdk.submitted_id in result.stdout
        assert fake_job_sdk.submitted_name in result.stdout

    def test_verbose_flag(self, fake_job_sdk, mock_job_controller):
        runner = CliRunner()
        runner.invoke(submit, ["-n", "custom-name", "--", "python", "main.py"])

        # No verbose flag -- exclude details
        result = runner.invoke(status, ["-n", "custom-name"])
        assert result.exit_code == 0
        assert "id" in result.stdout
        assert "name" in result.stdout
        assert "state" in result.stdout
        assert "config" not in result.stdout

        # Verbose flag -- include details
        result = runner.invoke(status, ["-n", "custom-name", "-v"])
        assert result.exit_code == 0
        assert "id" in result.stdout
        assert "name" in result.stdout
        assert "state" in result.stdout
        assert "config" in result.stdout

    def test_job_runs(self, fake_job_sdk, mock_job_controller):
        runner = CliRunner()
        runner.invoke(submit, ["--", "python", "main.py"])
        fake_job_sdk._job_runs = [
            JobRunStatus(name="failed-job-run", state=JobRunState.FAILED),
            JobRunStatus(name="succeeded-job-run", state=JobRunState.SUCCEEDED),
        ]
        result = runner.invoke(status, ["--job-id", fake_job_sdk.submitted_id])
        assert result.exit_code == 0
        assert (
            "runs:\n- name: failed-job-run\n  state: FAILED\n- name: succeeded-job-run\n  state: SUCCEEDED"
            in result.stdout
        )


class TestWait:
    def test_name(self, fake_job_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["-n", "custom-name"])
        assert result.exit_code == 0
        assert fake_job_sdk.waited_name == "custom-name"

    def test_id(self, fake_job_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["--id", "test-job-id"])
        assert result.exit_code == 0
        assert fake_job_sdk.waited_id == "test-job-id"

    def test_state(self, fake_job_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["-n", "custom-name", "--state", "FAILED"])
        assert result.exit_code == 0
        assert fake_job_sdk.waited_name == "custom-name"
        assert fake_job_sdk.waited_state == JobState.FAILED

    def test_bad_state(self, fake_job_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["-n", "custom-name", "--state", "FOOBAR"])
        assert result.exit_code == 1
        assert "'FOOBAR' is not a valid JobState" in str(result.stdout)

    def test_timeout(self, fake_job_sdk):
        runner = CliRunner()
        result = runner.invoke(wait, ["-n", "custom-name", "--timeout-s=30"])
        assert result.exit_code == 0
        assert fake_job_sdk.waited_timeout_s == 30



from typing import Optional

from rich.console import Console

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import (
    DeleteMachineRequest,
    ListMachinesResponse,
)
from anyscale.cloud import get_cloud_id_and_name
from anyscale.controllers.base_controller import BaseController


class MachineController(BaseController):
    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log
        self.console = Console()

    def list_machines(self, machine_pool_name: str,) -> ListMachinesResponse:
        response: ListMachinesResponse = self.api_client.list_machines_api_v2_machines_get(
            machine_pool_name=machine_pool_name,
        ).result
        return response

    def delete_machine(
        self,
        cloud: Optional[str],
        cloud_id: Optional[str],
        machine_id: str,
        machine_pool_name: Optional[str],
    ):
        if cloud is not None:
            cloud_id, _ = get_cloud_id_and_name(self.api_client, cloud_name=cloud)

        self.api_client.delete_machine_api_v2_machines_delete_post(
            DeleteMachineRequest(
                cloud_id=cloud_id,
                machine_id=machine_id,
                machine_pool_name=machine_pool_name,
            ),
        )



import dataclasses
from typing import Any, Dict, Optional

import click

from anyscale.authenticate import get_auth_api_client
from anyscale.cli_logger import BlockLogger
from anyscale.connect import ClientBuilder
from anyscale.connect_utils.prepare_cluster import create_prepare_cluster_block
from anyscale.connect_utils.project import create_project_block
from anyscale.util import PROJECT_NAME_ENV_VAR


log = BlockLogger()  # Anyscale CLI Logger


@dataclasses.dataclass
class ClusterInfo:
    """
    Synchronize with ray.dashboard.modules.job.sdk.ClusterInfo
    """

    address: str
    cookies: Optional[Dict[str, Any]] = None
    metadata: Optional[Dict[str, Any]] = None
    headers: Optional[Dict[str, Any]] = None


def get_job_submission_client_cluster_info(
    address: str,
    create_cluster_if_needed: Optional[bool] = None,
    cookies: Optional[Dict[str, Any]] = None,  # noqa: ARG001
    metadata: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, Any]] = None,  # noqa: ARG001
    **kwargs: Any,  # noqa: ARG001
) -> ClusterInfo:
    """
    Get address and cookies used for ray JobSubmissionClient.

    Args:
        address (str): Address of same form as ray.init address without
            anyscale:// prefix.
        create_cluster_if_needed (bool): Indicates whether the cluster
            of the address returned needs to be running. Raise an error
            if cluster is not running and this is False. Create a cluster
            if cluster is not running and this is True.

    Returns:
        ClusterInfo object consisting of address, cookies, and metadata for
            JobSubmissionClient to use.
    """

    if create_cluster_if_needed:
        # Use ClientBuilder to start cluster if needed because cluster needs to be active for
        # the calling command.
        api_client = get_auth_api_client().api_client
        client_builder = ClientBuilder(
            address=address, log=BlockLogger(log_output=False)
        )
        project_block = create_project_block(
            client_builder._project_dir,  # noqa: SLF001
            client_builder._project_name,  # noqa: SLF001
            cloud_name=client_builder._cloud_name,  # noqa: SLF001
            cluster_compute_name=client_builder._cluster_compute_name,  # noqa: SLF001
            cluster_compute_dict=client_builder._cluster_compute_dict,  # noqa: SLF001
        )
        project_id = project_block.project_id
        project_dir = project_block.project_dir
        if project_dir and not client_builder._project_name:  # noqa: SLF001
            # Warning when project dir provided or found in current directory and project name not provided
            # as input.
            # TODO(nikita): Remove after .anyscale.yaml is no longer supported
            log.warning(
                f"Project directory {project_dir} was detected. Using a project directory "
                "to set the project for the cluster to use with `ray job submit` has been deprecated, and this "
                "functionality will be removed in April 2022. To start an Anyscale cluster "
                "in a particular project, please instead specify the project name as "
                f'RAY_ADDRESS="anyscale://{project_block.project_name}/{client_builder._cluster_name if client_builder._cluster_name else ""}". '  # noqa: SLF001
                "The project name can also be specified by setting the environment variable "
                f'{PROJECT_NAME_ENV_VAR}="{project_block.project_name}". '
                "Otherwise the cluster will not be grouped to a particular project.\n"
            )
        prepare_cluster_block = create_prepare_cluster_block(
            project_id=project_id,
            cluster_name=client_builder._cluster_name,  # noqa: SLF001
            autosuspend_timeout=client_builder._autosuspend_timeout,  # noqa: SLF001
            allow_public_internet_traffic=client_builder._allow_public_internet_traffic,  # noqa: SLF001
            needs_update=client_builder._needs_update,  # noqa: SLF001
            cluster_compute_name=client_builder._cluster_compute_name,  # noqa: SLF001
            cluster_compute_dict=client_builder._cluster_compute_dict,  # noqa: SLF001
            cloud_name=client_builder._cloud_name,  # noqa: SLF001
            build_pr=client_builder._build_pr,  # noqa: SLF001
            force_rebuild=client_builder._force_rebuild,  # noqa: SLF001
            build_commit=client_builder._build_commit,  # noqa: SLF001
            cluster_env_name=client_builder._cluster_env_name,  # noqa: SLF001
            cluster_env_dict=client_builder._cluster_env_dict,  # noqa: SLF001
            cluster_env_revision=client_builder._cluster_env_revision,  # noqa: SLF001
            ray=client_builder._ray,  # noqa: SLF001
        )
        cluster_name = prepare_cluster_block.cluster_name
    else:
        # Calling ClientBuilder to parse address.
        api_client = get_auth_api_client(log_output=False).api_client
        client_builder = ClientBuilder(
            address=address, log=BlockLogger(log_output=False)
        )
        cluster_name = client_builder._cluster_name  # noqa: SLF001
        project_block = create_project_block(
            client_builder._project_dir,  # noqa: SLF001
            client_builder._project_name,  # noqa: SLF001
            log_output=False,
        )
        project_id = project_block.project_id

    user = api_client.get_user_info_api_v2_userinfo_get().result
    metadata = {"creator_id": user.id}

    cluster_list = api_client.list_sessions_api_v2_sessions_get(
        project_id=project_id, active_only=True, name=cluster_name
    ).results

    if len(cluster_list) > 0:
        cluster = cluster_list[0]
        if cluster.host_name and cluster.access_token:
            return ClusterInfo(
                address=cluster.host_name,
                cookies={"anyscale-token": cluster.access_token},
                metadata=metadata,
            )
        else:
            raise click.ClickException(
                f"Host name or access token not found for cluster {cluster_name}. Please check the cluster is currently running."
            )
    else:
        raise click.ClickException(
            f"No running cluster found with name {cluster_name} in project {project_id}. Please start "
            "the cluster, or change the project context if the wrong one is being used."
        )



from typing import Tuple

import pytest

from anyscale._private.anyscale_client.fake_anyscale_client import FakeAnyscaleClient
from anyscale._private.models.image_uri import ImageURI
from anyscale.image._private.image_sdk import ImageSDK


@pytest.fixture()
def sdk_with_fake_client() -> Tuple[ImageSDK, FakeAnyscaleClient]:
    fake_client = FakeAnyscaleClient()
    return ImageSDK(client=fake_client), fake_client


class TestBuildImageFromRequirements:
    def test_build_image_from_requirements(
        self, sdk_with_fake_client: Tuple[ImageSDK, FakeAnyscaleClient],
    ):
        sdk, fake_client = sdk_with_fake_client
        build_id = "mybaseimageid"
        image_uri = ImageURI.from_str("docker.io/fake/default:latest")
        image_name = "bldname123"
        fake_client.set_image_uri_mapping(
            image_uri, build_id,
        )

        expected_containerfile = "\n".join(
            [
                "# syntax=docker/dockerfile:1",
                f"FROM {image_uri}",
                'RUN pip install "requests"',
                'RUN pip install "flask"',
            ]
        )

        fake_client.set_containerfile_mapping(
            expected_containerfile, build_id="my_new_build"
        )

        actual_build_id = sdk.build_image_from_requirements(
            image_name, build_id, requirements=["requests", "flask"],
        )

        assert actual_build_id == "my_new_build"



import re

from anyscale.sdk.anyscale_client.models.cluster_environment import ClusterEnvironment
from anyscale.sdk.anyscale_client.models.cluster_environment_build import (
    ClusterEnvironmentBuild,
)


IMAGE_URI_PATTERN = "[registry_host/]user_name/repository[:tag][@digest]"
IMAGE_URI_PATTERN_RE = re.compile(
    r"^"
    # Optional registry host: hostname with optional port
    r"((?P<host>[a-zA-Z0-9.-]+)(?::(?P<port>[0-9]+))?/)?"
    # Repository: user_name/repository
    r"(?P<user>[a-zA-Z0-9-_]+/)?"  # user is optional.
    r"(?P<repository>[a-zA-Z0-9-/_.]+)"
    # Optional Tag: version or string after ':'
    r"(:(?P<tag>[a-zA-Z0-9_.-]+))?"
    # Optional Digest: string after '@'
    # Note that when both tag and digest are provided, the tag is ignored.
    r"(@(?P<digest>[a-zA-Z0-9:]+))?"
    r"$"
)

ANYSCALE_CLUSTER_ENV_PREFIX = "anyscale/image/"


class ImageURI:
    image_uri: str

    @classmethod
    def from_str(cls, image_uri_str: str):
        if not isinstance(image_uri_str, str):
            raise TypeError("'image_uri' must be a string.")

        if image_uri_str == "":
            raise ValueError("image_uri cannot be empty.")

        matches = IMAGE_URI_PATTERN_RE.match(image_uri_str)
        if not matches:
            raise ValueError(
                f"Invalid image URI: '{image_uri_str}'. Must be in the format: '{IMAGE_URI_PATTERN}'."
            )
        instance = super().__new__(cls)
        instance.image_uri = image_uri_str
        return instance

    @classmethod
    def from_cluster_env_build(
        cls, cluster_env: ClusterEnvironment, build: ClusterEnvironmentBuild
    ):
        image_uri_str = (
            ANYSCALE_CLUSTER_ENV_PREFIX + f"{cluster_env.name}:{build.revision}"
        )
        instance = super().__new__(cls)
        instance.image_uri = image_uri_str
        return instance

    def to_cluster_env_name(self) -> str:
        """Convert the image URI to a cluster environment name."""
        pattern = re.compile("^[A-Za-z0-9_-]+$")
        # Keep only characters that match the pattern
        escaped = []
        for c in self.image_uri:
            if not pattern.match(c):
                escaped.append("-")
            else:
                escaped.append(c)
        return "".join(escaped)

    def is_cluster_env_image(self) -> bool:
        """Check if the image URI is a cluster environment image."""
        return self.image_uri.startswith(ANYSCALE_CLUSTER_ENV_PREFIX)

    def to_cluster_env_identifier(self) -> str:
        assert self.is_cluster_env_image()
        identifier = self.image_uri.split("/")[2]
        return identifier

    def __str__(self):
        return self.image_uri

    def __eq__(self, other):
        if isinstance(other, ImageURI):
            return self.image_uri == other.image_uri
        return False

    def __hash__(self):
        return hash(self.image_uri)



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
from enum import Enum, unique
import random
import string


@unique
class IDTypes(Enum):
    projects = "prj"
    snapshots = "sna"
    users = "usr"
    organizations = "org"
    organization_invitations = "orginv"
    organizations_marketing_questions = "orgmq"
    sessions = "ses"
    applied_snapshots = "asn"
    clouds = "cld"
    cloud_resources = "cldrsrc"
    cloud_resources_gcp = "cldrsrcgcp"
    permissions_projects = "ppr"
    permissions_sessions = "pse"
    permissions_snapshots = "psn"
    permissions_clouds = "pcl"
    permissions_organizations = "por"
    identities = "ide"
    autosync_sessions = "aus"
    session_state_transitions = "sst"
    server_sessions = "sss"
    session_commands = "scd"
    instances = "ins"
    instances_unique = "asi"
    application_templates = "apt"
    builds = "bld"
    session_operations = "sop"
    compute_templates = "cpt"
    usage_snapshots = "usp"
    usage_export_jobs = "uej"
    namespaces = "nsp"
    runtime_environments = "rte"
    actors = "act"
    jobs = "job"
    ray_job_submissions = "rjs"
    ha_job_notifications = "hjn"
    anonymous_jobs = "ajb"
    anonymous_actors = "aac"
    sso_configs = "sso"
    serve_deployments = "srvdep"
    ha_jobs = "prodjob"
    job_queues = "jq"
    service = "service"
    ha_job_configs = "hjc"
    ha_job_state = "hjs"
    anyscale_customer_zone = "acz"
    ha_job_goal_state = "hjgs"
    audit_log = "alg"
    cluster_statuses = "cstat"
    recent_activities = "ract"
    support_requests = "supr"
    experimental_workspaces = "expwrk"
    structured_outputs = "output"
    experimental_integrations = "expint"
    cron_job = "cronjob"
    cluster_startups = "csu"
    integrations = "int"
    service2 = "service2"
    service2_event = "service2event"
    service2_rollout = "srr"
    service2_version = "srv"
    one_time_password = "otp"  # pragma: allowlist secret
    cluster_events = "cevt"
    files = "file"
    organization_payment_info = "org_pi"
    aioa_cloud_waitlist = "aioa_cl_wlst"

    # Anyscale Endpoints tables
    endpoints_users = "euser"
    # ID for auth tokens for reference (primary key)
    endpoints_auth_tokens = "eauth"
    # Token for auth tokens to be shown to users. This token is only shown to users once.
    endpoints_auth_token_tokens = "esecret"
    # PK for auth sessions
    endpoints_sessions = "esession"
    # Token for auth sessions
    endpoints_session_token = "esessiont"
    # PK for magic links
    endpoints_magic_links = "eotp"
    # User-facing token for magic link
    endpoints_magic_link_tokens = "eotpt"
    # Metronome transaction ID
    endpoints_metronome_transactions = "mett"
    # Cluster usage ID
    cluster_usage = "cuse"
    # Fine-tuned models
    endpoints_fine_tuned_models = "efine"
    # AICAEndpoint
    endpoints_aica_endpoints = "aept"
    # AICAModel
    endpoints_aica_models = "amdl"
    # AICAModelDeployment
    endpoints_aica_model_deployments = "amdpl"
    # Fine-tune jobs
    endpoints_fine_tuning_jobs = "eftjob"
    # Fine-tune job config overrides
    endpoints_fine_tuning_job_overrides = "eftovd"
    # AICAEndpoint events
    endpoints_aica_endpoints_events = "aept_evt"
    # Endpoints vercel integrations
    endpoints_vercel_integrations = "evi"
    # Endpoints vercel integrations projects
    endpoints_vercel_integration_projects = "evp"
    # Anyscale Machine bug reports
    anyscale_machine_bug_reports = "ambr"
    # Serve deployment events
    serve_deployment_events = "srvdep_evt"
    email_magic_links = "eml"
    email_magic_link_tokens = "emlt"
    billing_versions = "billv"
    customer_aggregated_logs_configs = "calc"
    service_accounts = "sa"


_default_id_length: int = 26


def generate_token_str(length: int = _default_id_length) -> str:
    """Generates a token comprised of uppercase (letters), lowercase, and numbers (62 possible characters)."""
    return "".join(
        # `random.SystemRandom().choices()` is cryptographically secure
        # (as opposed to `random.choices()`)
        random.SystemRandom().choices(string.ascii_letters + string.digits, k=length)
    )


class IDGenerator:
    def __init__(self, type: IDTypes):  # noqa: A002
        self.type = type.value

        self.rand = random.SystemRandom()
        self.possible_characters = list(
            set(string.ascii_lowercase + string.digits)
            - {"o", "0"}  # No confusing characters
        )

    def generate_id(self, k: int = _default_id_length) -> str:
        id_part = "".join(self.rand.choices(self.possible_characters, k=k))
        return f"{self.type}_{id_part}"


# Use this to avoid global variables due to cloudpickle not having great support for them.
def generate_id(type: IDTypes, k: int = _default_id_length) -> str:  # noqa: A002
    generator = IDGenerator(type)
    return generator.generate_id(k=k)
